"""Contains history class and helper functions."""

import json
import pickle

from skorch.utils import open_file_like


# pylint: disable=invalid-name
class _none:
    """Special placeholder since ``None`` is a valid value."""


def _not_none(items):
    """Whether the item is a placeholder or contains a placeholder."""
    if not isinstance(items, (tuple, list)):
        items = (items,)
    return all(item is not _none for item in items)


def _getitem_list_list(items, keys, tuple_=False):
    """Ugly but efficient extraction of multiple values from a list of
    items.
    """
    filtered = []
    for item in items:
        row = []
        for key in keys:
            try:
                row.append(item[key])
            except KeyError:
                break
        else:  # no break
            if row:
                filtered.append(tuple(row) if tuple_ else row)
    if items and not filtered:
        return _none
    return filtered


def _getitem_list_tuple(items, keys):
    return _getitem_list_list(items, keys, tuple_=True)


def _getitem_list_str(items, key):
    filtered = []
    for item in items:
        try:
            filtered.append(item[key])
        except KeyError:
            continue
    if items and not filtered:
        return _none
    return filtered


def _getitem_dict_list(item, keys):
    return [item.get(key, _none) for key in keys]


def _getitem_dict_tuple(item, keys):
    return tuple(item.get(key, _none) for key in keys)


def _getitem_dict_str(item, key):
    return item.get(key, _none)


def _get_getitem_method(items, key):
    """Return method to extract values from items.

    For the given type of items and type of keys, find the correct
    method to extract the values. By calling this only once per items,
    we can save a lot of type checking, which can be slow if there are
    a lot of epochs and a lot of items. However, we now make the
    assumption that the type of items doesn't change (we know that the
    key doesn't change). This should always be true, except if
    something really weird happens.

    We are multi-dispatching based on the following possibilities:

    * history[0, 'foo', :10]: get a list of items
    * history[0, 'foo', 0]: get a dict
    * history[0, 'foo', :, 'bar']: key is a str
    * history[0, 'foo', :, ('bar', 'baz')]: key is list/tuple of str

    """
    if isinstance(items, list):
        if isinstance(key, list):
            return _getitem_list_list
        if isinstance(key, tuple):
            return _getitem_list_tuple
        if isinstance(key, str):
            return _getitem_list_str
        raise TypeError("History access with given types not supported")

    if isinstance(items, dict):
        if isinstance(key, list):
            return _getitem_dict_list
        if isinstance(key, tuple):
            return _getitem_dict_tuple
        if isinstance(key, str):
            return _getitem_dict_str
    raise TypeError("History access with given types not supported")


def _unpack_index(i):
    """Unpack index and return exactly four elements.

    If index is more shallow than 4, return None for trailing
    dimensions. If index is deeper than 4, raise a KeyError.

    """
    if len(i) > 4:
        raise KeyError(
            "Tried to index history with {} indices but only "
            "4 indices are possible.".format(len(i)))

    # fill trailing indices with None
    i_e, k_e, i_b, k_b = i + tuple([None] * (4 - len(i)))

    return i_e, k_e, i_b, k_b


class History(list):
    """History contains the information about the training history of
    a :class:`.NeuralNet`, facilitating some of the more common tasks
    that are occur during training.

    When you want to log certain information during training (say, a
    particular score or the norm of the gradients), you should write
    them to the net's history object.

    It is basically a list of dicts for each epoch, that, again,
    contains a list of dicts for each batch. For convenience, it has
    enhanced slicing notation and some methods to write new items.

    To access items from history, you may pass a tuple of up to four
    items:

      1. Slices along the epochs.
      2. Selects columns from history epochs, may be a single one or a
         tuple of column names.
      3. Slices along the batches.
      4. Selects columns from history batchs, may be a single one or a
         tuple of column names.

    You may use a combination of the four items.

    If you select columns that are not present in all epochs/batches,
    only those epochs/batches are chosen that contain said columns. If
    this set is empty, a ``KeyError`` is raised.

    Examples
    --------
    >>> # ACCESSING ITEMS
    >>> # history of a fitted neural net
    >>> history = net.history
    >>> # get current epoch, a dict
    >>> history[-1]
    >>> # get train losses from all epochs, a list of floats
    >>> history[:, 'train_loss']
    >>> # get train and valid losses from all epochs, a list of tuples
    >>> history[:, ('train_loss', 'valid_loss')]
    >>> # get current batches, a list of dicts
    >>> history[-1, 'batches']
    >>> # get latest batch, a dict
    >>> history[-1, 'batches', -1]
    >>> # get train losses from current batch, a list of floats
    >>> history[-1, 'batches', :, 'train_loss']
    >>> # get train and valid losses from current batch, a list of tuples
    >>> history[-1, 'batches', :, ('train_loss', 'valid_loss')]

    >>> # WRITING ITEMS
    >>> # add new epoch row
    >>> history.new_epoch()
    >>> # add an entry to current epoch
    >>> history.record('my-score', 123)
    >>> # add a batch row to the current epoch
    >>> history.new_batch()
    >>> # add an entry to the current batch
    >>> history.record_batch('my-batch-score', 456)
    >>> # overwrite entry of current batch
    >>> history.record_batch('my-batch-score', 789)

    """

    def new_epoch(self):
        """Register a new epoch row."""
        self.append({'batches': []})

    def new_batch(self):
        """Register a new batch row for the current epoch."""
        # pylint: disable=invalid-sequence-index
        self[-1]['batches'].append({})

    def record(self, attr, value):
        """Add a new value to the given column for the current
        epoch.

        """
        msg = "Call new_epoch before recording for the first time."
        if not self:
            raise ValueError(msg)
        self[-1][attr] = value

    def record_batch(self, attr, value):
        """Add a new value to the given column for the current
        batch.

        """
        # pylint: disable=invalid-sequence-index
        self[-1]['batches'][-1][attr] = value

    def to_list(self):
        """Return history object as a list."""
        return list(self)

    @classmethod
    def from_file(cls, f):
        """Load the history of a ``NeuralNet`` from a json file.

        Parameters
        ----------
        f : file-like object or str

        """

        with open_file_like(f, 'r') as fp:
            return cls(json.load(fp))

    def to_file(self, f):
        """Saves the history as a json file. In order to use this feature,
        the history must only contain JSON encodable Python data structures.
        Numpy and PyTorch types should not be in the history.

        Parameters
        ----------
        f : file-like object or str

        """
        with open_file_like(f, 'w') as fp:
            json.dump(self.to_list(), fp)

    def __getitem__(self, i):
        # This implementation resolves indexing backwards,
        # i.e. starting from the batches, then progressing to the
        # epochs.
        if isinstance(i, (int, slice)):
            i = (i,)

        # i_e: index epoch, k_e: key epoch
        # i_b: index batch, k_b: key batch
        i_e, k_e, i_b, k_b = _unpack_index(i)
        keyerror_msg = "Key {!r} was not found in history."

        if i_b is not None and k_e != 'batches':
            raise KeyError("History indexing beyond the 2nd level is "
                           "only possible if key 'batches' is used, "
                           "found key {!r}.".format(k_e))

        items = self.to_list()

        # extract the epochs
        # handles: history[i_e]
        if i_e is not None:
            items = items[i_e]
            if isinstance(i_e, int):
                items = [items]

        # extract indices of batches
        # handles: history[..., k_e, i_b]
        if i_b is not None:
            items = [row[k_e][i_b] for row in items]

        # extract keys of epochs or batches
        # handles: history[..., k_e]
        # handles: history[..., ..., ..., k_b]
        if k_e is not None and (i_b is None or k_b is not None):
            key = k_e if k_b is None else k_b

            if items:
                extract = _get_getitem_method(items[0], key)
                items = [extract(item, key) for item in items]

                # filter out epochs with missing keys
                items = list(filter(_not_none, items))

            if not items and not (k_e == 'batches' and i_b is None):
                # none of the epochs matched
                raise KeyError(keyerror_msg.format(key))

            if (
                    isinstance(i_b, slice)
                    and k_b is not None
                    and not any(batches for batches in items)
            ):
                # none of the batches matched
                raise KeyError(keyerror_msg.format(key))

        if isinstance(i_e, int):
            items, = items

        return items


class DistributedHistory:
    """History for use in training using multiple processes

    When using skorch with :class:`.AccelerateMixin` for multi GPU training, use
    this class instead of the default :class:`.History` class.

    When using PyTorch :class:`torch.nn.parallel.DistributedDataParallel`, the
    whole training process is forked and batches are processed in parallel. That
    means that the standard :class:`.History` does not see all the batches that
    are being processed, which results in the different processes having
    histories that are out of sync. This is bad because the history is used as a
    reference to influence the training, e.g. to control early stopping.

    This class solves the problem by using a distributed store from PyTorch,
    e.g. :class:`torch.distributed.TCPStore`, to synchronize the batch
    information across processes. This ensures that the information stored in
    the individual history copies is identical for ``history[:, 'batches']``.
    When it comes to the epoch-level information, it can still diverge between
    processes (e.g. the recorded duration of the epoch).

    To use this class, instantiate it and pass it as the ``history`` argument to
    the net.

    Click here for more information on `PyTorch distributed key-value stores
    <https://pytorch.org/docs/stable/distributed.html#distributed-key-value-store>`_.

    Notes
    -----
    If using this class results in the processes hanging or timing out, double
    check that the ``rank`` and ``world_size`` arguments are set correctly.
    Otherwise, the history instances will be waiting for records that are
    actually never written.

    If the speed of the processes is very uneven, there can also be timeouts. To
    increase the waiting time, pass a corresponding ``timeout`` argument to the
    ``store`` instance.

    Objects stored with this history make a json roundtrip. Therefore, if you
    store objects that don't survive a json roundtrip (say, numpy arrays), don't
    use this class.

    The PyTorch ``Store`` classes cannot be pickled. Therefore, if a net using
    this history class is pickled, the ``store`` attribute is discarded so that
    the pickling does not fail. This means, however, that an unpickled net
    cannot be used for further training without manually setting the ``store``
    attribute on the ``history``.

    Examples
    --------
    >>> # general
    >>> from skorch import NeuralNetClassifier
    >>> from torch.distributed import TCPStore
    >>> from torch.nn.parallel import DistributedDataParallel
    >>> def train(rank, world_size, is_master):
    ...     store = TCPStore(
    ...         "127.0.0.1", port=1234, world_size=world_size)
    ...     dist_history = DistributedHistory(
    ...         store=store, rank=rank, world_size=world_size)
    ...     net = NeuralNetClassifier(..., history=dist_history)
    ...     net.fit(X, y)
    >>> # with accelerate
    >>> from accelerate import Accelerator
    >>> from skorch.hf import AccelerateMixin
    >>> accelerator = Accelerator(...)
    >>> def train(accelerator):
    ...     is_master = accelerator.is_main_process
    ...     world_size = accelerator.num_processes
    ...     rank = accelerator.local_process_index
    ...     store = TCPStore(
    ...         "127.0.0.1", port=1234, world_size=world_size, is_master=is_master)
    ...     dist_history = DistributedHistory(
    ...         store=store, rank=rank, world_size=world_size)
    ...     net = AcceleratedNet(..., history=dist_history)
    ...     net.fit(X, y)

    Parameters
    ----------
    store : torch.distributed.Store
      The torch distributed ``Store`` instance,
      :class:`torch.distributed.TCPStore` has been tested to work.

    rank : int
      The rank of this particular process among all processes. Each process
      should have a unique rank between 0 and ``world_size`` - 1. If using
      ``accelerate``, the rank can be determined as
      ``accelerator.local_process_index``.

    world_size : int
      The number of processes in the training. When using ``accelerate``, the
      world size can be determined as ``accelerator.num_processes``.

    Attributes
    ----------
    history : skorch.history.History
      The actual skorch ``History`` object can be accessed using the
      :class:`.History` attribute. You should call ``net.history.sync()`` to
      ensure that all data is synced into the history before reading from it.

    """
    def __init__(self, *args, store, rank, world_size):
        self.store = store
        self.rank = rank
        self.world_size = world_size

        self._history = History(*args)
        self._cur_batch = 0
        self._sync_queue = []

    @property
    def history(self):
        self.sync()
        return self._history

    def to_list(self):
        return self.history.to_list()

    @classmethod
    def from_file(cls, f):
        raise NotImplementedError

    def to_file(self, f):
        return self.history.to_file(f)

    def __len__(self):
        return len(self.history)

    def __delitem__(self, i):
        raise NotImplementedError

    def clear(self):
        self.history.clear()

    def new_epoch(self):
        self.sync()
        self._history.new_epoch()
        self._cur_batch = 0

    def new_batch(self):
        self.sync()
        self._history.new_batch()
        self._cur_batch += 1

    def record(self, attr, value):
        self._history.record(attr, value)

    def record_batch(self, attr, value):
        """Add a new value to the given column for the current
        batch.

        Instead of writing to the history directly, write to the distributed
        store. Then, once the history is being read from, the values from the
        store are synchronized.

        This class "remembers" which values were written to the store by
        creating a key that uniquely identifies the values. Choosing a correct
        key is crucial here, since it must not only be unique (say, a uuid), but
        also sufficient to replay the history so that they can be recorded
        correctly.

        When the structure of the key is changed, it must be changed accordingly
        inside of the sync method.

        """
        rank = self.rank
        batch = self._cur_batch
        epoch = len(self._history)
        key = [epoch, batch, rank, attr]
        self.store.set(json.dumps(key), json.dumps(value))

        # the queue not only stores the keys for this process, but the keys for
        # all processes
        for rank in range(self.world_size):
            key = [epoch, batch, rank, attr]
            self._sync_queue.append(key)

    def __getitem__(self, i):
        self.sync()
        return self._history[i]

    def sync(self):
        """Collect batch records across all ranks from store and write them to
        the history

        Syncing is not a single atomic operation, if something breaks the flow,
        we can end up with an inconsistent state.

        """
        if not self._sync_queue:
            return

        # this ensures that that the values are visited in a stable order
        # grouped by batches: order by batch count and within the batches in the
        # order of the rank and within the ranks in the alphabetical order of
        # the attributes
        keys_sorted = sorted(self._sync_queue)
        keys_json = [json.dumps(key) for key in keys_sorted]
        self.store.wait(keys_json)

        batch_prev = None
        rank_prev = None

        for (_, batch, rank, attr), key_json in zip(keys_sorted, keys_json):
            if batch_prev is None:
                batch_prev = batch
            if rank_prev is None:
                rank_prev = rank

            # each time the batch counter or rank changes, it means that the
            # next batch needs to be started
            if (batch != batch_prev) or (rank != rank_prev):
                self._history.new_batch()
                batch_prev = batch
                rank_prev = rank

            val = json.loads(self.store.get(key_json))
            self._history.record_batch(attr, val)

        self._sync_queue.clear()

    def __getstate__(self):
        state = self.__dict__.copy()
        try:
            # Unfortunately, TCPStore and FileStore are not pickleable. We still
            # try, in case the user provides another type that can be pickled.
            pickle.dumps(state['store'])
        except (TypeError, pickle.PicklingError):
            # TCPStore and FileStore raise TypeError, others could be
            # PicklingError
            state['store'] = None
        return state



"""Tests for history.py."""

import pickle
import time
import multiprocessing as mp
from functools import partial

import numpy as np
import pytest
import torch
from torch.distributed import TCPStore
from sklearn.datasets import make_classification


class TestHistory:

    test_epochs = 3
    test_batches = 4

    @pytest.fixture(scope='class', params=['single', 'distributed'])
    def history_cls(self, request):
        # run tests once with default History, once with DistributedHistory
        from skorch.history import DistributedHistory, History
        from skorch._version import Version

        if request.param == 'single':
            return History

        if request.param == 'distributed':
            # TODO: remove this once PyTorch 1.11 is no longer supported
            version_torch = Version(torch.__version__)
            if version_torch < Version('1.12.0'):
                # Tests with DistributedHistory are flaky with PyTorch 1.11 and when
                # using pytest on CI. Running directly, i.e. without pytest, does
                # not appear to cause issues.
                pytest.skip(msg="Skipping multiprocessing for PyTorch < 1.12")

            store = TCPStore("127.0.0.1", port=1234, world_size=1, is_master=True)
            return partial(
                DistributedHistory, store=store, rank=0, world_size=1
            )

        raise ValueError("Incorrect pytest request parameter '{request.param}'")

    @pytest.fixture
    def history(self, history_cls):
        """Return a history filled with epoch and batch data."""
        h = history_cls()
        for num_epoch in range(self.test_epochs):
            h.new_epoch()
            h.record('duration', 1)
            h.record('total_loss', num_epoch + self.test_batches)
            if num_epoch == 2:
                h.record('extra', 42)
            for num_batch in range(self.test_batches):
                h.new_batch()
                h.record_batch('loss', num_epoch + num_batch)
                if num_batch % 2 == 0 and (num_epoch + 1) != self.test_epochs:
                    h.record_batch('extra_batch', 23)
        return h

    @pytest.fixture
    def ref(self, history):
        return history.to_list()

    def test_list_initialization(self, history_cls):
        h = history_cls([1, 2, 3])
        assert len(h) == 3

    def test_history_length(self, history):
        assert len(history) == self.test_epochs
        # we expect to have the extracted batches for each epoch
        assert len(history[:, 'batches']) == self.test_epochs

    def test_history_epoch_column(self, history, ref):
        total_losses = history[:, 'total_loss']
        total_losses_ref = [n['total_loss'] for n in ref]
        assert total_losses == total_losses_ref

    def test_history_epoch_two_columns(self, history, ref):
        duration_with_losses = history[:, ('total_loss', 'duration')]

        total_losses_ref = [n['total_loss'] for n in ref]
        durations_ref = [n['duration'] for n in ref]
        expected = list(zip(total_losses_ref, durations_ref))

        assert duration_with_losses == expected

    def test_history_epoch_two_columns_different_order(self, history, ref):
        duration_with_losses = history[:, ('duration', 'total_loss')]

        total_losses_ref = [n['total_loss'] for n in ref]
        durations_ref = [n['duration'] for n in ref]
        expected = list(zip(durations_ref, total_losses_ref))

        assert duration_with_losses == expected

    def test_history_partial_index(self, history, ref):
        extra = history[:, 'extra']
        assert len(extra) == 1
        # we retrieve 'extra' from a slice, therefore we expect a list as result
        assert extra == [ref[2]['extra']]

    def test_history_partial_and_full_index(self, history, ref):
        total_loss_with_extra = history[:, ('total_loss', 'extra')]

        assert len(total_loss_with_extra) == 1
        assert total_loss_with_extra[0][0] == ref[2]['total_loss']
        assert total_loss_with_extra[0][1] == ref[2]['extra']

    def test_history_partial_join_list(self, history, ref):
        total = history[:, ['total_loss', 'extra', 'batches']]

        # there's only one epoch with the 'extra' key.
        assert len(total) == 1
        assert total[0][0] == ref[2]['total_loss']
        assert total[0][1] == ref[2]['extra']
        assert total[0][2] == ref[2]['batches']

    def test_history_retrieve_single_value(self, history, ref):
        total_loss_0 = history[0, 'total_loss']
        assert total_loss_0 == ref[0]['total_loss']

    def test_history_retrieve_multiple_values(self, history, ref):
        total_loss_0_to_1 = history[0:1, 'total_loss']
        assert total_loss_0_to_1 == [n['total_loss'] for n in ref[0:1]]

    def test_history_non_existing_values(self, history):
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            history[:, 'non-existing']
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            history[0, 'extra']

    def test_history_non_existing_values_batch(self, history):
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            history[:, 'batches', :, 'non-existing']
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            history[:, 'batches', 1, 'extra_batch']

    def test_history_mixed_slicing(self, history, ref):
        losses = history[:, 'batches', 0, 'loss']

        assert len(losses) == self.test_epochs
        assert losses == [epoch['batches'][0]['loss'] for epoch in ref]

        losses = history[0, 'batches', :, 'loss']
        assert losses == [batch['loss'] for batch in ref[0]['batches']]

    def test_history_partial_and_full_index_batches(self, history, ref):
        loss_with_extra = history[:, 'batches', :, ('loss', 'extra_batch')]

        expected_e0 = [(b['loss'], b['extra_batch']) for b in ref[0]['batches']
                       if 'extra_batch' in b]
        expected_e1 = [(b['loss'], b['extra_batch']) for b in ref[1]['batches']
                       if 'extra_batch' in b]

        assert len(loss_with_extra) == self.test_epochs - 1
        assert loss_with_extra[0] == expected_e0
        assert loss_with_extra[1] == expected_e1

    def test_history_partial_batches_batch_key_3rd(self, history, ref):
        extra_batches = history[:, 'batches', :, 'extra_batch']

        expected_e0 = [b['extra_batch'] for b in ref[0]['batches']
                       if 'extra_batch' in b]
        expected_e1 = [b['extra_batch'] for b in ref[1]['batches']
                       if 'extra_batch' in b]

        # In every epoch there are 2 batches with the 'extra_batch'
        # key except for the last epoch. We therefore two results
        # of which one of them is an empty list.
        assert len(extra_batches) == self.test_epochs - 1
        assert extra_batches[0] == expected_e0
        assert extra_batches[1] == expected_e1

    def test_history_partial_batches_batch_key_4th(self, history, ref):
        extra_batches = history[:, 'batches', :, 'extra_batch']

        expected_e0 = [b['extra_batch'] for b in ref[0]['batches']
                       if 'extra_batch' in b]
        expected_e1 = [b['extra_batch'] for b in ref[1]['batches']
                       if 'extra_batch' in b]

        # In every epoch there are 2 batches with the 'extra_batch'
        # key except for the last epoch. We therefore two results
        # of which one of them is an empty list.
        assert len(extra_batches) == self.test_epochs - 1
        assert extra_batches[0] == expected_e0
        assert extra_batches[1] == expected_e1

    def test_history_partial_singular_values(self, history):
        values = history[-1, ('duration', 'total_loss')]
        expected = (history[-1]['duration'], history[-1]['total_loss'])

        # pylint: disable=unidiomatic-typecheck
        assert type(values) == tuple
        assert values == expected

    def test_history_slice_beyond_batches_but_key_not_batches(self, history):
        with pytest.raises(KeyError) as exc:
            # pylint: disable=pointless-statement
            history[:, 'not-batches', 0]

        msg = exc.value.args[0]
        expected = ("History indexing beyond the 2nd level is "
                    "only possible if key 'batches' is used, "
                    "found key 'not-batches'.")
        assert msg == expected

    def test_history_with_invalid_epoch_key(self, history):
        key = slice(None), 'not-batches'
        with pytest.raises(KeyError) as exc:
            # pylint: disable=pointless-statement
            history[key]

        msg = exc.value.args[0]
        expected = "Key 'not-batches' was not found in history."
        assert msg == expected

    def test_history_too_many_indices(self, history):
        with pytest.raises(KeyError) as exc:
            # pylint: disable=pointless-statement
            history[:, 'batches', :, 'train_loss', :]

        msg = exc.value.args[0]
        expected = ("Tried to index history with 5 indices but only "
                    "4 indices are possible.")
        assert msg == expected

    def test_history_save_load_cycle_file_obj(self, history_cls, history, tmpdir):
        if hasattr(history, 'store'):
            # DistributedHistory does not support loading from file
            pytest.skip()

        history_f = tmpdir.mkdir('skorch').join('history.json')

        with open(str(history_f), 'w') as f:
            history.to_file(f)

        with open(str(history_f), 'r') as f:
            new_history = history_cls.from_file(f)

        assert history == new_history

    def test_history_save_load_cycle_file_path(self, history_cls, history, tmpdir):
        if hasattr(history, 'store'):
            # DistributedHistory does not support loading from file
            pytest.skip()

        history_f = tmpdir.mkdir('skorch').join('history.json')

        history.to_file(str(history_f))
        new_history = history_cls.from_file(str(history_f))

        assert history == new_history

    @pytest.mark.parametrize('type_', [list, tuple])
    def test_history_multiple_keys(self, history, type_):
        dur_loss = history[-1, type_(['duration', 'total_loss'])]
        # pylint: disable=unidiomatic-typecheck
        assert type(dur_loss) is type_ and len(dur_loss) == 2

        loss_loss = history[-1, 'batches', -1, type_(['loss', 'loss'])]
        # pylint: disable=unidiomatic-typecheck
        assert type(loss_loss) is type_ and len(loss_loss) == 2

    def test_history_key_in_other_epoch(self, history_cls):
        h = history_cls()
        for has_valid in (True, False):
            h.new_epoch()
            h.new_batch()
            h.record_batch('train_loss', 1)
            if has_valid:
                h.new_batch()
                h.record_batch('valid_loss', 2)

        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            h[-1, 'batches', -1, 'valid_loss']

    def test_history_no_epochs_index(self, history_cls):
        h = history_cls()
        with pytest.raises(IndexError):
            # pylint: disable=pointless-statement
            h[-1, 'batches']

    def test_history_jagged_batches(self, history_cls):
        h = history_cls()
        for num_batch in (1, 2):
            h.new_epoch()
            for _ in range(num_batch):
                h.new_batch()
        # Make sure we can access this batch
        assert h[-1, 'batches', 1] == {}

    @pytest.mark.parametrize('value, check_warn', [
        ([], False),
        (np.array([]), True),
    ])
    def test_history_retrieve_empty_list(self, value, history_cls, check_warn, recwarn):
        h = history_cls()
        if hasattr(h, 'store') and isinstance(value, np.ndarray):
            # DistributedHistory does not support numpy arrays because they
            # cannot be json serialized
            pytest.skip()

        h.new_epoch()
        h.record('foo', value)
        h.new_batch()
        h.record_batch('batch_foo', value)

        # Make sure we can access our object
        out = h[-1, 'foo']
        assert (out is value) or (out == value)
        out = h[-1, 'batches', -1, 'batch_foo']
        assert (out is value) or (out == value)

        # There should be no warning about comparison to an empty ndarray
        if check_warn:
            assert not recwarn.list

    @pytest.mark.parametrize('has_epoch, epoch_slice', [
        (False, slice(None)),
        (True, slice(1, None)),
    ])
    def test_history_no_epochs_key(self, has_epoch, epoch_slice, history_cls):
        h = history_cls()
        if has_epoch:
            h.new_epoch()

        # Expect KeyError since the key was not found in any epochs
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            h[epoch_slice, 'foo']
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            h[epoch_slice, ['foo', 'bar']]

    @pytest.mark.parametrize('has_batch, batch_slice', [
        (False, slice(None)),
        (True, slice(1, None)),
    ])
    def test_history_no_batches_key(self, has_batch, batch_slice, history_cls):
        h = history_cls()
        h.new_epoch()
        if has_batch:
            h.new_batch()

        # Expect KeyError since the key was not found in any batches
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            h[-1, 'batches', batch_slice, 'foo']
        with pytest.raises(KeyError):
            # pylint: disable=pointless-statement
            h[-1, 'batches', batch_slice, ['foo', 'bar']]

    @pytest.mark.parametrize('has_epoch, epoch_slice', [
        (False, slice(None)),
        (True, slice(1, None)),
    ])
    def test_history_no_epochs_batches(self, has_epoch, epoch_slice, history_cls):
        h = history_cls()
        if has_epoch:
            h.new_epoch()

        # Expect a list of zero epochs since 'batches' always exists
        assert h[epoch_slice, 'batches'] == []
        assert h[epoch_slice, 'batches', -1] == []

    def test_pickle(self, history):
        loaded = pickle.loads(pickle.dumps(history))

        # The store cannot be pickled, so it is set to None
        if not hasattr(history, 'store'):
            assert history.to_list() == loaded.to_list()
        else:
            assert loaded.store is None


class TestDistributedHistoryMultiprocessing:
    """Testing DistributedHistory with multiprocessing

    This is not a proper test for how DistributedHistory should be used in the
    wild, which would be in a multi-GPU setting using DDP (possibly through
    accelerate). However, since we cannot test this setting in CI, this is the
    next best approximation.

    Since DDP creates forks of the main process, each one would have its own
    history instance, leading in them diverging. This can be bad, e.g. if the
    history is used to steer the training (think early stopping). This test
    intends to show that the histories do not diverge.

    In a sense, this test actually tests incorrect behavior because the
    different processes all write the exact same content to the history.
    However, if DDP was used, each process would see different batches and thus
    write different records to their histories. So when reading this test,
    imagine that this would be what's truly happening.

    This test is very high level, unlike TestHistory. For instance, there is no
    attempt at testing "weird" read and write patterns, like creating multiple
    empty batches. It is very much assumed that the history is used "correctly",
    as in a normal skorch fit loop.

    """
    @pytest.fixture
    def data(self, classifier_data):
        X, y = classifier_data
        return X[:100], y[:100]

    @staticmethod
    def train(args):
        from skorch import NeuralNetClassifier
        from skorch.history import DistributedHistory

        time.sleep(0.5)  # wait a bit, else broken pipe error may occur

        torch.manual_seed(0)
        rank, nprocs, module, (X, y) = args

        # let's hope the port is free
        store = TCPStore(
            '127.0.0.1', 8890 + nprocs, world_size=nprocs, is_master=(rank == 0)
        )
        dist_history = DistributedHistory(store=store, rank=rank, world_size=nprocs)
        net = NeuralNetClassifier(module, max_epochs=2, history=dist_history)
        net.history = dist_history
        net.fit(X, y)

        time.sleep(0.5)  # wait a bit, else broken pipe error may occur

        return net

    @pytest.mark.parametrize('nprocs', [1, 2, 3, 4])
    def test_distributed_history(self, nprocs):
        from skorch.toy import make_classifier
        from skorch._version import Version

        # TODO: remove this once PyTorch 1.11 is no longer supported
        version_torch = Version(torch.__version__)
        if version_torch < Version('1.12.0'):
            # Tests with DistributedHistory are flaky with PyTorch 1.11 and when
            # using pytest on CI. Running directly, i.e. without pytest, does
            # not appear to cause issues.
            pytest.skip(msg="Skipping multiprocessing for PyTorch < 1.12")

        X, y = make_classification(
            500, 20, n_informative=10, random_state=0, flip_y=0.1
        )
        X = X.astype(np.float32)
        y = y.astype(np.int64)

        torch.set_num_threads(1)
        module = make_classifier(
            input_units=20,
            hidden_units=10,
            num_hidden=2,
            dropout=0.5,
        )

        args = zip(
            range(nprocs), [nprocs] * nprocs, [module] * nprocs, [(X, y)] * nprocs
        )
        # https://github.com/pytorch/pytorch/wiki/Autograd-and-Fork
        ctx = mp.get_context('spawn')
        with ctx.Pool(nprocs) as pool:
            nets = pool.map(self.train, args)

        # expected entries for training batches:
        # - 500 samples, 400 used for train, 100 for valid
        # - batch size is 128, i.e. 4 batches per epoch for train
        # - the 100 valid samples fit into 1 batch
        # note that when using DDP correctly, since batches are split across the
        # processes, there would *not* be nprocs times more entries.
        train_batches = 4
        valid_batches = 1
        for net in nets:
            h = net.history
            assert len(h[0, 'batches', :, 'train_loss']) == train_batches * nprocs
            assert len(h[1, 'batches', :, 'train_loss']) == train_batches * nprocs
            assert len(h[0, 'batches', :, 'valid_loss']) == valid_batches * nprocs
            assert len(h[1, 'batches', :, 'valid_loss']) == valid_batches * nprocs

        # the batch information is synchronized, so should be identical; the
        # epoch information is *not* synchronized and can diverge, most notably
        # with things like the duration
        for net0, net1 in zip(nets, nets[1:]):
            train_loss0 = net0.history[:, 'batches', :, 'train_loss']
            train_loss1 = net1.history[:, 'batches', :, 'train_loss']
            assert train_loss0 == train_loss1

            valid_loss0 = net0.history[:, 'batches', :, 'valid_loss']
            valid_loss1 = net1.history[:, 'batches', :, 'valid_loss']
            assert valid_loss0 == valid_loss1

            assert net0.history[:, 'dur'] != net1.history[:, 'dur']



"""Tests for regressor.py

Only contains tests that are specific for regressor subclasses.

"""

from functools import partial

import numpy as np
import pytest
from sklearn.base import clone

from skorch.tests.conftest import INFERENCE_METHODS


class TestNeuralNetRegressor:
    @pytest.fixture(scope='module')
    def data(self, regression_data):
        return regression_data

    @pytest.fixture(scope='module')
    def module_cls(self):
        from skorch.toy import make_regressor
        return make_regressor(dropout=0.5)

    @pytest.fixture(scope='module')
    def module_pred_1d_cls(self):
        from skorch.toy import MLPModule
        # Module that returns 1d predictions
        return partial(MLPModule, output_units=1, squeeze_output=True)

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetRegressor
        return NeuralNetRegressor

    @pytest.fixture(scope='module')
    def net(self, net_cls, module_cls):
        return net_cls(
            module_cls,
            max_epochs=20,
            lr=0.1,
        )

    @pytest.fixture(scope='module')
    def multioutput_module_cls(self):
        from skorch.toy import make_regressor
        return make_regressor(output_units=3, dropout=0.5)

    @pytest.fixture(scope='module')
    def multioutput_net(self, net_cls, multioutput_module_cls):
        return net_cls(
            multioutput_module_cls,
            max_epochs=1,
            lr=0.1,
        )

    @pytest.fixture(scope='module')
    def net_fit(self, net, data):
        # Careful, don't call additional fits on this, since that would have
        # side effects on other tests.
        X, y = data
        return net.fit(X, y)

    def test_clone(self, net_fit):
        clone(net_fit)

    def test_fit(self, net_fit, recwarn):
        # fitting does not raise anything and does not warn
        assert not recwarn.list

    @pytest.mark.parametrize('method', INFERENCE_METHODS)
    def test_not_fitted_raises(self, net_cls, module_cls, data, method):
        from skorch.exceptions import NotInitializedError
        net = net_cls(module_cls)
        X = data[0]
        with pytest.raises(NotInitializedError) as exc:
            # we call `list` because `forward_iter` is lazy
            list(getattr(net, method)(X))

        msg = ("This NeuralNetRegressor instance is not initialized "
               "yet. Call 'initialize' or 'fit' with appropriate arguments "
               "before using this method.")
        assert exc.value.args[0] == msg

    def test_net_learns(self, net, net_cls, data, module_cls):
        X, y = data
        net = net_cls(
            module_cls,
            max_epochs=10,
            lr=0.1,
        )
        net.fit(X, y)
        train_losses = net.history[:, 'train_loss']
        assert train_losses[0] > 2 * train_losses[-1]

    def test_history_default_keys(self, net_fit):
        expected_keys = {'train_loss', 'valid_loss', 'epoch', 'dur', 'batches'}
        for row in net_fit.history:
            assert expected_keys.issubset(row)

    def test_predict_predict_proba(self, net_fit, data):
        X = data[0]
        y_pred = net_fit.predict(X)

        # predictions should not be all zeros
        assert not np.allclose(y_pred, 0)

        y_proba = net_fit.predict_proba(X)
        # predict and predict_proba should be identical for regression
        assert np.allclose(y_pred, y_proba, atol=1e-6)

    def test_score(self, net_fit, data):
        X, y = data
        r2_score = net_fit.score(X, y)
        assert r2_score <= 1.

    def test_multioutput_score(self, multioutput_net, multioutput_regression_data):
        X, y = multioutput_regression_data
        multioutput_net.fit(X, y)
        r2_score = multioutput_net.score(X, y)
        assert r2_score <= 1.

    def test_dimension_mismatch_warning(self, net_cls, module_cls, data, recwarn):
        # When the target and the prediction have different dimensionality, mse
        # loss will broadcast them, calculating all pairwise errors instead of
        # only sample-wise. Since the errors are averaged at the end, there is
        # still a valid loss, which makes the error hard to spot. Thankfully,
        # torch gives a warning in that case. We test that this warning exists,
        # otherwise, skorch users could run into very hard to debug issues
        # during training.
        net = net_cls(module_cls)
        X, y = data
        X, y = X[:100], y[:100].flatten()  # make y 1d
        net.fit(X, y)

        w0, w1 = recwarn.list  # one warning for train, one for valid 
        # The warning comes from PyTorch, so checking the exact wording is prone to
        # error in future PyTorch versions. We thus check a substring of the
        # whole message and cross our fingers that it's not changed.
        msg_substr = (
            "This will likely lead to incorrect results due to broadcasting. "
            "Please ensure they have the same size"
        )
        assert msg_substr in str(w0.message)
        assert msg_substr in str(w1.message)

    def test_fitting_with_1d_target_and_pred(
            self, net_cls, module_cls, data, module_pred_1d_cls, recwarn
    ):
        # This test relates to the previous one. In general, users should fit
        # with target and prediction being 2d, even if the 2nd dimension is just
        # 1. However, in some circumstances (like when using BaggingRegressor,
        # see next test), having the ability to fit with 1d is required. In that
        # case, the module output also needs to be 1d for correctness.
        X, y = data
        X, y = X[:100], y[:100]  # less data to run faster
        y = y.flatten()

        net = net_cls(module_pred_1d_cls)
        net.fit(X, y)
        assert not recwarn.list

    def test_bagging_regressor(
            self, net_cls, module_cls, data, module_pred_1d_cls, recwarn
    ):
        # https://github.com/skorch-dev/skorch/issues/972
        from sklearn.ensemble import BaggingRegressor

        net = net_cls(module_pred_1d_cls)  # module output should be 1d too
        X, y = data
        X, y = X[:100], y[:100]  # less data to run faster
        y = y.flatten()  # make y 1d or else sklearn will complain
        regr = BaggingRegressor(net, n_estimators=2, random_state=0)
        regr.fit(X, y)  # does not raise
        assert not recwarn.list  # ensure there is no broadcast warning from torch



"""Using LLMs from transformers for zero/few-shot learning

Open tasks:

- Have a "fast/greedy" option for predict - it is not necessary to calculate
  probabilities for all classes up until the last token. When class A has
  probability p_A and class B has p_B(t) < p_A at token t, then no matter what
  the probability for later tokens, it cannot surpass p_A anymore.

- A small use case where the classifiers are used as a transformer in a bigger
  pipeline, e.g. to extract structured knowledge from a text ("Does this
  product description contain the size of the item?")

- a way to format the text/labels/few-shot samples before they're
  string-interpolated, maybe Jinja2?

- Test if this works with a more diverse range of LLMs

- Enable multi-label classification. Would probably require sigmoid instead of
  softmax and a (empirically determined?) threshold.

- Check if it is possible to enable caching for encoder-decoder LLMs like
  flan-t5.

"""

import importlib
import warnings
from string import Formatter

import numpy as np
import torch
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.utils import check_random_state
from sklearn.utils.validation import check_is_fitted
from transformers import AutoConfig, AutoTokenizer, LogitsProcessor

from skorch.exceptions import LowProbabilityError
from skorch.llm.prompts import DEFAULT_PROMPT_FEW_SHOT, DEFAULT_PROMPT_ZERO_SHOT, DELIM


def _check_format_string(text, kwargs):
    """Check if string can be exactly formatted with kwargs

    This is necessary because even though calling .format(...) on a string will
    raise if a key is missing, it ignores extra keys, and we want to error in
    this case.

    Gives a warning if the text contains placeholders that are not in kwargs, or
    if kwargs contains keys that are not in the text.

    Parameters
    ----------
    text : str
      The string to format.

    kwargs : dict
      The values to use for formatting.

    """
    formatter = Formatter()
    keys = {key for _, key, _, _ in formatter.parse(text) if key is not None}
    keys_expected = set(kwargs.keys())
    num_keys = len(keys_expected)

    if keys == keys_expected:
        return

    keys_missing = keys_expected - keys
    keys_extra = keys - keys_expected
    msg = (
        f"The prompt may not be correct, it expects {num_keys} "
        "placeholders: " + ", ".join(f"'{key}'" for key in sorted(keys_expected))
    )
    if keys_missing:
        msg += ", missing keys: "
        msg += ", ".join(f"'{key}'" for key in sorted(keys_missing))
    if keys_extra:
        msg += ", extra keys: "
        msg += ", ".join(f"'{key}'" for key in sorted(keys_extra))
    warnings.warn(msg)


def _load_model_and_tokenizer(
        name, device, architectures=('Generation', 'LMHead', 'CausalLM')
):
    """Load a transformers model based only on its name

    This is a bit tricky, because we usually require the task as well to load
    the correct model. However, asking users to pass name and task is not very
    user friendly because it makes, for instance, defining grid search
    parameters more cumbersome if two parameters need to be changed together.

    To solve this, we basically guess the correct architecture based on its
    name.

    Parameters
    ----------
    name : str
      The name of the transformers model to load, e.g. ``google/flan-t5-small``.

    device : str or torch.device
      The device to use for the model, e.g. ``'cpu'`` or
      ``torch.device('cuda:0')``.

    architectures : tuple of str (default=('Generation', 'LMHead', 'CausalLM'))
      The architectures allowed to be loaded. An architecture is chosen if one
      of the strings is a substring of the actual architecture name.

    Returns
    -------
    model : torch.nn.Module
      The pre-trained transformers model.

    tokenizer
      The tokenizer for the model.

    Raises
    ------
    ValueError
      If the model architecture cannot be inferred from the name, raise a
      ``ValueError``.

    """
    config = AutoConfig.from_pretrained(name)

    architecture = None
    for arch in config.architectures:
        if any(substr in arch for substr in architectures):
            architecture = arch
            break

    if architecture is None:
        raise ValueError(
            f"Could not identify architecture for model '{name}', try loading "
            "model and tokenizer directly using the corresponding 'Auto' classes "
            "from transformers and pass them to the classifier"
        )

    transformers_module = importlib.import_module("transformers")
    cls = getattr(transformers_module, architecture, None)
    if cls is None:
        raise ValueError(
            f"Could not find a class '{architecture}' in transformers, try loading "
            "model and tokenizer directly using the corresponding 'Auto' classes "
            "from transformers and pass them to the classifier"
        )

    model = cls.from_pretrained(name).to(device)
    tokenizer = AutoTokenizer.from_pretrained(name)
    return model, tokenizer


def _extend_inputs(inputs, extra):
    """Extend input arguments with an extra column"""
    input_ids = inputs['input_ids']
    attention_mask = inputs['attention_mask']
    extra = torch.atleast_2d(
        torch.tensor(extra, dtype=torch.long, device=input_ids.device)
    )
    inputs_extended = inputs.copy()

    inputs_extended['input_ids'] = torch.cat((input_ids, extra), dim=-1)
    inputs_extended['attention_mask'] = torch.cat(
        (attention_mask, torch.ones_like(extra)), dim=-1
    )
    return inputs_extended


class _LogitsRecorder(LogitsProcessor):
    """Helper class to record logits and force the given label token ids"""
    def __init__(self, label_ids, tokenizer):
        self.recorded_scores = []
        self.label_ids = label_ids
        self.tokenizer = tokenizer

    def __call__(self, input_ids, scores):
        idx = len(self.recorded_scores)
        # we pull the logits to CPU because they are not used as input,
        # therefore there is no device mismatch and we save a bit of GPU memory
        self.recorded_scores.append(scores[0].clone().cpu())
        mask = torch.ones(scores.size(), dtype=torch.bool)
        mask[0, self.label_ids[idx]] = False
        scores[mask] = -float('inf')
        return scores


class _CacheModelWrapper:
    """Helper class that caches model generations

    For label ids, if one token sequence is [1, 2, 3] and the next token
    sequence is [1, 2, 4], for the 2nd sequence, the generation will retrieve
    the cached logits for [1, 2] and only generate [4].

    Set use_caching=False to disable it, e.g. for debugging.

    """
    def __init__(self, model, tokenizer, use_caching=True):
        self.model = model
        self.tokenizer = tokenizer
        self.use_caching = use_caching
        self.cache = {}
        self._total_calls = 0
        self._uncached_calls = 0

    def clear(self):
        self.cache.clear()

    def _make_key(self, kwargs):
        input_ids = kwargs['input_ids']
        input_id = input_ids[0].tolist()
        key = str(input_id)
        return key

    def get_cache(self, kwargs):
        if not self.use_caching:
            return

        key = self._make_key(kwargs)
        val = self.cache.get(key)
        return val

    def set_cache(self, kwargs, label_id, scores):
        if not self.use_caching:
            return

        key = self._make_key(kwargs)
        # store 1st element
        self.cache[key] = scores[0]

        # note that label_id i corresponds to score i+1
        # this is because the first score is for the input w/o label_id (only
        # the prompt); for this reason, the two sequences are offset by +1
        input_id = kwargs['input_ids'][0].tolist()
        for lid, score in zip(label_id, scores[1:]):
            input_id.append(lid)
            key = str(input_id)
            self.cache[key] = score

    def generate_logits(self, *, label_id, **kwargs):
        self._total_calls += 1  # mainly for debugging

        recorded_logits = []
        logits_cached = self.get_cache(kwargs)
        while logits_cached is not None:
            if label_id[0] == self.tokenizer.eos_token_id:
                # don't extend with eos_token -- it is already there at the end,
                # we don't need it twice
                break

            recorded_logits.append(logits_cached)
            kwargs = _extend_inputs(kwargs, label_id[:1])
            label_id = label_id[1:]
            logits_cached = self.get_cache(kwargs)

        if not label_id:
            # the whole generation was cached
            return recorded_logits

        if label_id[0] == self.tokenizer.pad_token_id:
            # no need to generate on pad tokens
            return recorded_logits

        self._uncached_calls += 1  # mainly for debugging
        recorder = _LogitsRecorder(
            label_ids=label_id,
            tokenizer=self.tokenizer,
        )
        self.model.generate(
            logits_processor=[recorder],
            # TODO: should this be the max len of all labels?
            max_new_tokens=len(label_id),
            **kwargs
        )
        self.set_cache(kwargs, label_id, recorder.recorded_scores)
        return recorded_logits + recorder.recorded_scores[:]


class _LlmBase(BaseEstimator, ClassifierMixin):
    """Base class for LLM models

    This class handles a few of the checks, as well as the whole prediction
    machinery.

    Required attributes are:

    - model_name
    - model
    - tokenizer
    - prompt
    - probas_sum_to_1
    - device
    - error_low_prob
    - threshold_low_prob
    - use_caching

    """
    def check_X_y(self, X, y, **fit_params):
        raise NotImplementedError

    def check_prompt(self, prompt):
        raise NotImplementedError

    def get_prompt(self, text):
        raise NotImplementedError

    def fit(self, X, y, **fit_params):
        # note: should call _fit
        raise NotImplementedError

    def check_classes(self, y):
        return np.unique(y)

    def check_is_fitted(self):
        required_attrs = ['model_', 'tokenizer_', 'prompt_', 'classes_', 'label_ids_']
        check_is_fitted(self, required_attrs)

    @property
    def device_(self):
        # Use whatever device the model is on, not self.device. If a user
        # initializes with a model that is on GPU, we don't want to require them
        # to adjust the device argument, which would be annoying
        return self.model_.device

    def check_args(self):
        # users should either pass the model name, or the model and tokenizer,
        # but not both
        cls_name = self.__class__.__name__
        msg = (
            f"{cls_name} needs to be initialized with either a model name, "
            "or a model & tokenizer, but not both."
        )
        if self.model_name is not None:
            if (self.model is not None) or (self.tokenizer is not None):
                raise ValueError(msg)
        else:
            if (self.model is None) or (self.tokenizer is None):
                raise ValueError(msg)

        possible_values_error_low_prob = ['ignore', 'raise', 'warn', 'return_none']
        if self.error_low_prob not in possible_values_error_low_prob:
            raise ValueError(
                "error_low_prob must be one of "
                f"{', '.join(possible_values_error_low_prob)}; "
                f"got {self.error_low_prob} instead"
            )

        if (self.threshold_low_prob < 0) or (self.threshold_low_prob > 1):
            raise ValueError(
                "threshold_low_prob must be between 0 and 1, "
                f"got {self.threshold_low_prob} instead"
            )

    def _is_encoder_decoder(self, model):
        return hasattr(model, 'get_encoder')

    def _fit(self, X, y, **fit_params):
        """Prepare everything to enable predictions."""
        self.check_args()
        self.check_X_y(X, y)
        if self.model_name is not None:
            self.model_, self.tokenizer_ = _load_model_and_tokenizer(
                self.model_name, device=self.device
            )
        else:
            self.model_, self.tokenizer_ = self.model, self.tokenizer

        if self._is_encoder_decoder(self.model_) and self.use_caching:
            # Explanation: When we have a prompt [1, 2, 3] and label [4, 5], and
            # if the model is an encoder-decoder architecture (seq2seq), then
            # [1, 2, 3] is encoded and [4, 5] are generated by the decoder. If
            # we wanted to cache, we would store the result (among others) for
            # [1, 2, 3, 4]. However, encoding [1, 2, 3, 4] and generating [5] is
            # not the same operation as encoding [1, 2, 3] and generating [4,
            # 5]. Granted, the logits could be very close, but we cannot be sure
            # and it will depend on the model. Therefore, for the time being, we
            # don't allow caching for encoder-decoder models.
            raise ValueError(
                "Caching is not supported for encoder-decoder models, "
                "initialize the model with use_caching=False."
            )

        self.classes_ = self.check_classes(y)
        self.prompt_ = self.check_prompt(self.prompt)
        classes = [str(c) for c in self.classes_]
        self.label_ids_ = self.tokenizer_(classes)['input_ids']
        self.cached_model_ = _CacheModelWrapper(
            self.model_, self.tokenizer_, use_caching=self.use_caching
        )
        return self

    def _predict_one(self, text):
        """Make a prediction for a single sample

        The returned probabilites are *not normalized* yet.

        Raises a ``LowProbabilityError`` if the total probability of all labels
        is 0, or, assuming ``error_low_prob`` is ``'raise'``, when it is below
        ``threshold_low_prob``. This check is done here because we want to raise
        eagerly instead of going through all samples and then raise.

        """
        inputs = self.tokenizer_(text, return_tensors='pt').to(self.device_)

        probas_all_labels = []
        for label_id in self.label_ids_:
            logits = self.cached_model_.generate_logits(label_id=label_id, **inputs)
            logits = torch.vstack(logits)
            probas = torch.nn.functional.softmax(logits, dim=-1)

            # in case we deal with BFloats which are not supported on CPU
            probas = probas.to(torch.float)

            n = len(logits)
            label_id = label_id[:n]  # don't need EOS, PAD, etc.
            probas_at_idx = probas[torch.arange(n), label_id]
            # multiplying the probabilities for token A, B, C, ... This is
            # because we are interested in P(A) AND P(B|A) AND P(C|A,B), etc.
            # p(A) * P(B|A) =
            # p(A) * p(AnB) / p(A) =
            # p(A) * p(A) * p(B) / p(A) =
            # p(A) * p(B)
            # thus the product of individual probalities is correct
            probas_all_labels.append(torch.prod(probas_at_idx).item())

        prob_sum = sum(probas_all_labels)
        if prob_sum == 0.0:
            raise LowProbabilityError("The sum of all probabilities is zero.")

        probs_are_low = prob_sum < self.threshold_low_prob
        if probs_are_low and (self.error_low_prob == 'raise'):
            raise LowProbabilityError(
                f"The sum of all probabilities is {prob_sum:.3f}, "
                f"which is below the minimum threshold of {self.threshold_low_prob:.3f}"
            )

        y_prob = np.array(probas_all_labels).astype(np.float64)
        return y_prob

    def _predict_proba(self, X):
        """Return the unnormalized y_proba

        Warns if the total probability for a sample is below the threshold and
        ``error_low_prob`` is ``'warn'``.

        """
        self.check_is_fitted()
        y_proba = []
        for xi in X:
            text = self.get_prompt(xi)
            proba = self._predict_one(text)
            y_proba.append(proba)
        y_proba = np.vstack(y_proba)

        if self.error_low_prob == 'warn':
            total_low_probas = (y_proba.sum(1) < self.threshold_low_prob).sum()
            if total_low_probas:
                warnings.warn(
                    f"Found {total_low_probas} samples to have a total probability "
                    f"below the threshold of {self.threshold_low_prob:.3f}."
                )

        return y_proba

    def predict_proba(self, X):
        """Return the probabilities predicted by the LLM.

        Predictions will be forced to be one of the labels the model learned
        during ``fit``. Each column in ``y_proba`` corresponds to one class. The
        order of the classes can be checked in the ``.classes_`` attribute. In
        general, it is alphabetic. So the first column is the probability for
        the first class in ``.classes_``, the second column is the probability
        for the second class in ``.classes_``, etc.

        If ``error_low_prob`` is set to ``'warn'``, then a warning is given if,
        for at least one sample, the sum of the probabilities for all classes is
        below the threshold set in ``threshold_low_prob``. If ``error_low_prob``
        is set to ``'raise'``, then an error is raised instead.

        Parameters
        ----------
        X : input data
          Typically, this is a list/array of strings. E.g., this could be a list
          of reviews and the target is the sentiment. Technically, however, this
          can also contain numerical or categorical data, although it is
          unlikely that the LLM will generate good predictions for those.

        Returns
        -------
        y_proba : numpy ndarray
          The probabilities for each class. If ``probas_sum_to_1`` is set to
          ``False``, then the sum of the probabilities for each sample will not
          add up to 1

        """
        # y_proba not normalized here
        y_proba = self._predict_proba(X)

        if self.probas_sum_to_1:
            # normalizing here is okay because we already checked earlier that
            # the sum is not 0
            y_proba /= y_proba.sum(1, keepdims=True)

        return y_proba

    def predict(self, X):
        """Return the classes predicted by the LLM.

        Predictions will be forced to be one of the labels the model learned
        during ``fit`` (see the ``classes_`` attribute). The model will never
        predict a different label.

        If ``error_low_prob`` is set to ``'warn'``, then a warning is given if,
        for at least one sample, the sum of the probabilities for all classes is
        below the threshold set in ``threshold_low_prob``. If ``error_low_prob``
        is set to ``'raise'``, then an error is raised instead.

        If ``error_low_prob`` is set to ``'return_none'``, then the predicted
        class will be replaced by ``None`` if the sum of the probabilities is
        below the threshold set in ``threshold_low_prob``.

        Parameters
        ----------
        X : input data
          Typically, this is a list/array of strings. E.g., this could be a list
          of reviews and the target is the sentiment. Technically, however, this
          can also contain numerical or categorical data, although it is
          unlikely that the LLM will generate good predictions for those.

        Returns
        -------
        y_pred : numpy ndarray
          The label for each class.

        """
        # y_proba not normalized but it's not neeeded here
        y_proba = self._predict_proba(X)
        pred_ids = y_proba.argmax(1)
        y_pred = self.classes_[pred_ids]

        if self.error_low_prob == 'return_none':
            y_pred = y_pred.astype(object)  # prevents None from being cast to str
            mask_low_probas = y_proba.sum(1) < self.threshold_low_prob
            y_pred[mask_low_probas] = None
        return y_pred

    def clear_model_cache(self):
        """Clear the cache of the model

        Call this to free some memory.

        """
        if self.use_caching and hasattr(self, 'cached_model_'):
            # caching is used and model cache has been initialized
            self.cached_model_.clear()

    def __repr__(self):
        # The issue is that the repr for the transformer model can be quite huge
        # and the repr for the tokenizer quite ugly. We solve this by
        # temporarily replacing them by their class names. We should, however,
        # not completely rewrite the repr code because sklearn does some special
        # magic with the repr. This is why super().__repr__() should be used.
        if (self.model is None) and (self.tokenizer is None):
            # the user initialized the class with model name, the repr should be
            # fine
            return super().__repr__()

        model_prev = self.model
        tokenizer_prev = self.tokenizer
        rep = None

        # we are very defensive, making sure to restore the temporarily made
        # change in case of an error
        try:
            self.model = self.model.__class__.__name__
            self.tokenizer = self.tokenizer.__class__.__name__
            rep = super().__repr__()
        finally:
            self.model = model_prev
            self.tokenizer = tokenizer_prev

        if rep is None:
            # if the repr could not be generated, fall back to using the ugly
            # repr
            rep = super().__repr__()

        return rep


class ZeroShotClassifier(_LlmBase):
    """Zero-shot classification using a Large Language Model (LLM).

    This class allows you to use an LLM from Hugging Face transformers for
    zero-shot classification. There is no training during the ``fit`` call,
    instead, the LLM will be prompted to predict the labels for each sample.

    Parameters
    ----------
    model_name : str or None (default=None)
      The name of the model to use. This is the same name as used on Hugging
      Face Hub. For example, to use GPT2, pass ``'gpt2'``, to use the small
      flan-t5 model, pass ``'google/flan-t5-small'``. If the ``model_name``
      parameter is passed, don't pass ``model`` or ``tokenizer`` parameters.

    model : torch.nn.Module or None (default=None)
      The model to use. This should be a PyTorch text generation model from
      Hugging Face Hub or a model with the same API. Most notably, the model
      should have a ``generate`` method. If you pass the ``model``, you should
      also pass the ``tokenizer``, but you shall not pass the ``model_name``.

      Passing the model explicitly instead of the ``model_name`` can have a few
      advantages. Most notably, this allows you to modify the model, e.g.
      changing its config or how the model is loaded. For instance, some models
      can only be loaded with the option ``trust_remote_code=True``. If using
      the ``model_name`` argument, the default settings will be used instead.
      Passing the model explicitly also allows you to use custom models that are
      not uploaded to Hugging Face Hub.

    tokenizer (default=None)
      A tokenizer that is compatible with the model. Typically, this is loaded
      using the ``AutoTokenizer.from_pretrained`` method provided by Hugging
      Face transformers. If you pass the ``tokenizer``, you should also pass the
      ``model``, but you should not pass the ``model_name``.

    prompt : str or None (default=None)

      The prompt to use. This is the text that will be passed to the model to
      generate the prediction. If no prompt is passed, a default prompt will be
      used. The prompt should be a Python string with two placeholders, one
      called ``text`` and one called ``labels``. The ``text`` placeholder will
      replaced by the contents from ``X`` and the ``labels`` placeholder will be
      replaced by the unique labels taken from ``y``.

      An example prompt could be something like this:

          "Classify this text: {text}. Possible labels are {labels}". Your
          response: "

      All general tips for good prompt crafting apply here as well. Be aware
      that if the prompt is too long, it will exceed the context size of the
      model.

    probas_sum_to_1 : bool (default=True)
      If ``True``, then the probabilities for each sample will be normalized to
      sum to 1. If ``False``, the probabilities will not be normalized.

      In general, without normalization, the probabilities will not sum to 1
      because the LLM can generate any token, not just the labels. Since the
      model is restricted to only generate the available labels, there will be
      some probability mass that is unaccounted for. You could consider the
      missing probability mass to be an implicit 'other' class.

      In general, you should set this parameter to ``True`` because the default
      assumption is that probabilities sum to 1. However, setting this to
      ``False`` can be useful for debugging purposes, as it allows you to see
      how much probability the LLM assigns to different tokens. If the total
      probabilities are very low, it could be a sign that the LLM is not
      powerful enough or that the prompt is not well crafted.

    device : str or torch.device (default='cpu')
      The device to use. In general, using a GPU or other accelerated hardware
      is advised if runtime performance is critical.

      Note that if the ``model`` parameter is passed explicitly, the device of
      that model takes precedence over the value of ``device``.

    error_low_prob : {'ignore', 'warn', 'raise', 'return_none'} (default='ignore')
      Controls what should happen if the sum of the probabilities for a sample
      is below a given threshold. When encountering low probabilities, the
      options are to do one of the following:

        - ``'ignore'``: do nothing
        - ``'warn'``: issue a warning
        - ``'raise'``: raise an error
        - ``'return_none'``: return ``None`` as the prediction when calling
          ``.predict``

      The threshold is controlled by the ``threshold_low_prob`` parameter.

    threshold_low_prob : float (default=0.0)
      The threshold for the sum of the probabilities below which they are
      considered to be too low. The consequences of low probabilities are
      controlled by the ``error_low_prob`` parameter.

    use_caching : bool (default=True)
      If ``True``, the predictions for each sample will be cached, as well as
      the intermediate result for each generated token. This can speed up
      predictions when some samples are duplicated, or when labels have a long
      common prefix. An example of the latter would be if a label is called
      "intent.support.email" and another label is called "intent.support.phone",
      then the tokens for the common prefix "intent.support." are reused for
      both labels, as their probabilities are identical.

      Note that caching is currently not supported for encoder-decoder
      architectures such as flan-t5. If you want to use such an architecture,
      turn caching off.

      If you see any issues you might suspect are caused by caching, turn this
      option off, see if it helps, and report the issue on the skorch GitHub
      page.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes, )
      A list of class labels known to the classifier. This attribute can be used
      to identify which column in the probabilties returned by ``predict_proba``
      corresponds to which class.

    """
    def __init__(
            self,
            model_name=None,
            *,
            model=None,
            tokenizer=None,
            prompt=None,
            probas_sum_to_1=True,
            device='cpu',
            error_low_prob='ignore',
            threshold_low_prob=0.0,
            use_caching=True,
    ):
        self.model_name = model_name
        self.model = model
        self.tokenizer = tokenizer
        self.prompt = prompt
        self.probas_sum_to_1 = probas_sum_to_1
        self.device = device
        self.error_low_prob = error_low_prob
        self.threshold_low_prob = threshold_low_prob
        self.use_caching = use_caching

    def check_prompt(self, prompt):
        """Check if the prompt is well formed.

        If no prompt is provided, return the default prompt.

        Raises
        ------
        ValueError
          When the prompt is not well formed.

        """
        if prompt is None:
            prompt = DEFAULT_PROMPT_ZERO_SHOT


        kwargs = {
            'text': "some text",
            'labels': ["foo", "bar"],
        }
        _check_format_string(prompt, kwargs)
        return prompt

    def get_prompt(self, text):
        """Return the prompt for the given sample."""
        self.check_is_fitted()
        return self.prompt_.format(text=text, labels=self.classes_.tolist())

    def check_X_y(self, X, y, **fit_params):
        """Check that input data is well-behaved."""
        # X can be None but not y
        if y is None:
            raise ValueError(
                "y cannot be None, as it is used to infer the existing classes"
            )

        if not isinstance(y[0], str):
            # don't raise an error, as, hypothetically, the LLM could also
            # predict encoded targets, but it's not advisable
            warnings.warn(
                "y should contain the name of the labels as strings, e.g. "
                "'positive' and 'negative', don't pass label-encoded targets"
            )

    def fit(self, X, y, **fit_params):
        """Prepare everything to enable predictions.

        There is no actual fitting going on here, as the LLM is used as is.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
          The input data. For zero-shot classification, this can be ``None``.

        y : array-like of shape (n_samples,)
          The target classes. Ensure that each class that the LLM should be able
          to predict is present at least once. Classes that are not present
          during the ``fit`` call will never be predicted.

        **fit_params : dict
          Additional fitting parameters. This is mostly a placeholder for
          sklearn-compatibility, as there is no actual fitting process.

        Returns
        -------
        self
          The fitted estimator.

        """
        return self._fit(X, y, **fit_params)


class FewShotClassifier(_LlmBase):
    """Few-shot classification using a Large Language Model (LLM).

    This class allows you to use an LLM from Hugging Face transformers for
    few-shot classification. There is no training during the ``fit`` call,
    instead, the LLM will be prompted to predict the labels for each sample.

    Parameters
    ----------
    model_name : str or None (default=None)
      The name of the model to use. This is the same name as used on Hugging
      Face Hub. For example, to use GPT2, pass ``'gpt2'``, to use the small
      flan-t5 model, pass ``'google/flan-t5-small'``. If the ``model_name``
      parameter is passed, don't pass ``model`` or ``tokenizer`` parameters.

    model : torch.nn.Module or None (default=None)
      The model to use. This should be a PyTorch text generation model from
      Hugging Face Hub or a model with the same API. Most notably, the model
      should have a ``generate`` method. If you pass the ``model``, you should
      also pass the ``tokenizer``, but you shall not pass the ``model_name``.

      Passing the model explicitly instead of the ``model_name`` can have a few
      advantages. Most notably, this allows you to modify the model, e.g.
      changing its config or how the model is loaded. For instance, some models
      can only be loaded with the option ``trust_remote_code=True``. If using
      the ``model_name`` argument, the default settings will be used instead.
      Passing the model explicitly also allows you to use custom models that are
      not uploaded to Hugging Face Hub.

    tokenizer (default=None)
      A tokenizer that is compatible with the model. Typically, this is loaded
      using the ``AutoTokenizer.from_pretrained`` method provided by Hugging
      Face transformers. If you pass the ``tokenizer``, you should also pass the
      ``model``, but you should not pass the ``model_name``.

    prompt : str or None (default=None)

      The prompt to use. This is the text that will be passed to the model to
      generate the prediction. If no prompt is passed, a default prompt will be
      used. The prompt should be a Python string with three placeholders, one
      called ``text``, one called ``labels``, and one called ``examples``. The
      ``text`` placeholder will replaced by the contents from ``X`` passed to
      during inference and the ``labels`` placeholder will be replaced by the
      unique labels taken from ``y``. The examples will be taken from the ``X``
      and ``y`` seen during ``fit``.

      An example prompt could be something like this:

          "Classify this text: {text}. Possible labels are: {labels}. Here are
          some examples: {examples}. Your response: ".

      All general tips for good prompt crafting apply here as
      well. Be aware that if the prompt is too long, it will exceed the context
      size of the model.

    probas_sum_to_1 : bool (default=True)
      If ``True``, then the probabilities for each sample will be normalized to
      sum to 1. If ``False``, the probabilities will not be normalized.

      In general, without normalization, the probabilities will not sum to 1
      because the LLM can generate any token, not just the labels. Since the
      model is restricted to only generate the available labels, there will be
      some probability mass that is unaccounted for. You could consider the
      missing probability mass to be an implicit 'other' class.

      In general, you should set this parameter to ``True`` because the default
      assumption is that probabilities sum to 1. However, setting this to
      ``False`` can be useful for debugging purposes, as it allows you to see
      how much probability the LLM assigns to different tokens. If the total
      probabilities are very low, it could be a sign that the LLM is not
      powerful enough or that the prompt is not well crafted.

    max_samples: int (default=5)
      The number of samples to use for few-shot learning. The few-shot samples
      are taken from the ``X`` and ``y`` passed to ``fit``.

      This number should be large enough for the LLM to generalize, but not too
      large so as to exceed the context window size. More samples will also
      lower prediction speed.

    device : str or torch.device (default='cpu')
      The device to use. In general, using a GPU or other accelerated hardware
      is advised if runtime performance is critical.

      Note that if the ``model`` parameter is passed explicitly, the device of
      that model takes precedence over the value of ``device``.

    error_low_prob : {'ignore', 'warn', 'raise', 'return_none'} (default='ignore')
      Controls what should happen if the sum of the probabilities for a sample
      is below a given threshold. When encountering low probabilities, the
      options are to do one of the following:

        - ``'ignore'``: do nothing
        - ``'warn'``: issue a warning
        - ``'raise'``: raise an error
        - ``'return_none'``: return ``None`` as the prediction when calling
          ``.predict``

      The threshold is controlled by the ``threshold_low_prob`` parameter.

    threshold_low_prob : float (default=0.0)
      The threshold for the sum of the probabilities below which they are
      considered to be too low. The consequences of low probabilities are
      controlled by the ``error_low_prob`` parameter.

    use_caching : bool (default=True)
      If ``True``, the predictions for each sample will be cached, as well as
      the intermediate result for each generated token. This can speed up
      predictions when some samples are duplicated, or when labels have a long
      common prefix. An example of the latter would be if a label is called
      "intent.support.email" and another label is called "intent.support.phone",
      then the tokens for the common prefix "intent.support." are reused for
      both labels, as their probabilities are identical.

      Note that caching is currently not supported for encoder-decoder
      architectures such as flan-t5. If you want to use such an architecture,
      turn caching off.

      If you see any issues you might suspect are caused by caching, turn this
      option off, see if it helps, and report the issue on the skorch GitHub
      page.

    random_state : int, RandomState instance or None (default=None)
      The choice of examples that are picked for few-shot learning is random. To
      fix the random seed, use this argument.

    Attributes
    ----------
    classes_ : ndarray of shape (n_classes, )
      A list of class labels known to the classifier. This attribute can be used
      to identify which column in the probabilties returned by ``predict_proba``
      corresponds to which class.

    """
    def __init__(
            self,
            model_name=None,
            *,
            model=None,
            tokenizer=None,
            prompt=None,
            probas_sum_to_1=True,
            max_samples=5,
            device='cpu',
            error_low_prob='ignore',
            threshold_low_prob=0.0,
            use_caching=True,
            random_state=None,
    ):
        self.model_name = model_name
        self.model = model
        self.tokenizer = tokenizer
        self.prompt = prompt
        self.probas_sum_to_1 = probas_sum_to_1
        self.max_samples = max_samples
        self.device = device
        self.error_low_prob = error_low_prob
        self.threshold_low_prob = threshold_low_prob
        self.use_caching = use_caching
        self.random_state = random_state

    def check_prompt(self, prompt):
        """Check if the prompt is well formed.

        If no prompt is provided, return the default prompt.

        Raises
        ------
        ValueError
          When the prompt is not well formed.

        """

        if prompt is None:
            prompt = DEFAULT_PROMPT_FEW_SHOT

        kwargs = {
            'text': "some text",
            'labels': ["foo", "bar"],
            'examples': ["some examples"],
        }
        _check_format_string(prompt, kwargs)
        return prompt

    def get_examples(self, X, y, n_samples):
        """Given input data ``X`` and ``y``, return a subset of ``n_samples``
        for few-shot learning.

        This method aims at providing at least one example for each existing
        class.

        """
        examples = []
        seen_X = set()
        seen_y = set()
        rng = check_random_state(self.random_state)
        indices = rng.permutation(np.arange(len(y)))

        # first batch, fill with one example for each label
        for i in indices:
            if y[i] not in seen_y:
                examples.append((X[i], y[i]))
                seen_X.add(str(X[i]))
                seen_y.add(y[i])
            if len(seen_y) == len(self.classes_):
                # each label is represented
                break
            if len(examples) == n_samples:
                break

        if len(examples) == n_samples:
            return examples

        # second batch, fill with random other examples
        for i in indices:
            if str(X[i]) in seen_X:
                continue

            examples.append((X[i], y[i]))
            if len(examples) == n_samples:
                break

        # return in reverse order so that the label diversity from the 1st batch
        # comes last
        return examples[::-1]


    def fit(self, X, y, **fit_params):
        """Prepare everything to enable predictions.

        There is no actual fitting going on here, as the LLM is used as is. The
        examples used for few-shot learning will be derived from the provided
        input data. The selection mechanism for this is that, for each possible
        label, at least one example is taken from the data (if ``max_samples``
        is large enough).

        To change the way that examples are selected, override the
        ``get_examples`` method.

        Parameters
        ----------
        X : array-like of shape (n_samples,)
          The input data. For zero-shot classification, this can be ``None``.

        y : array-like of shape (n_samples,)
          The target classes. Ensure that each class that the LLM should be able
          to predict is present at least once. Classes that are not present
          during the ``fit`` call will never be predicted.

        **fit_params : dict
          Additional fitting parameters. This is mostly a placeholder for
          sklearn-compatibility, as there is no actual fitting process.

        Returns
        -------
        self
          The fitted estimator.

        """
        self._fit(X, y, **fit_params)
        n_samples = min(self.max_samples, len(y))
        self.examples_ = self.get_examples(X, y, n_samples=n_samples)
        return self

    def get_prompt(self, text):
        """Return the prompt for the given sample."""
        self.check_is_fitted()
        few_shot_examples = []
        for xi, yi in self.examples_:
            few_shot_examples.append(
                f"{DELIM}\n{xi}\n{DELIM}\n\nYour response:\n{yi}\n"
            )
        examples = "\n".join(few_shot_examples)
        return self.prompt_.format(
            text=text, labels=self.classes_.tolist(), examples=examples
        )

    def check_X_y(self, X, y, **fit_params):
        """Check that input data is well-behaved."""
        if (X is None) or (not len(X)):
            raise ValueError("For few-shot learning, pass at least one example")

        if y is None:
            raise ValueError(
                "y cannot be None, as it is used to infer the existing classes"
            )

        if not isinstance(y[0], str):
            # don't raise an error, as, hypothetically, the LLM could also
            # predict encoded targets, but it's not advisable
            warnings.warn(
                "y should contain the name of the labels as strings, e.g. "
                "'positive' and 'negative', don't pass label-encoded targets"
            )

        len_X, len_y = len(X), len(y)
        if len_X != len_y:
            raise ValueError(
                "X and y don't have the same number of samples, found "
                f"{len_X} and {len_y} samples, respectively"
        )



"""Different prompts used for LLMs

Note: When changing the prompt, some tests may start to fail because they check
the predictions using the default prompts.

"""


DELIM = "```"

DEFAULT_PROMPT_ZERO_SHOT = f"""You are a text classification assistant.

The text to classify:

{DELIM}
{{text}}
{DELIM}

Choose the label among the following possibilities with the highest probability.
Only return the label, nothing more:

{{labels}}

Your response:
"""

DEFAULT_PROMPT_FEW_SHOT = f"""You are a text classification assistant.

Choose the label among the following possibilities with the highest probability.
Only return the label, nothing more:

{{labels}}

Here are a few examples:

{{examples}}

The text to classify:

{DELIM}
{{text}}
{DELIM}

Your response:
"""



"""Provides helper class SkorchDoctor that assists understanding and debugging
the neural net training

"""

from collections.abc import Mapping
from functools import partial

import numpy as np
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted as sk_check_is_fitted
import torch

from skorch.exceptions import NotInitializedError
from skorch.utils import to_numpy


def named_modules(net):
    """For a skorch net, yield all module names and modules

    Ignore modules that are not ``torch.nn.Module`` instances.

    """
    # pylint: disable=protected-access
    for module_name in net._modules + net._criteria:
        module = getattr(net, module_name + '_')
        if isinstance(module, torch.nn.Module):
            yield module_name, module


# pylint: disable=unused-argument,redefined-builtin
def _add_activation_hook(model, input, output, *, recs, layer_name):
    """Helper function for adding activation hooks"""

    # only record training data
    if not model.training:
        return

    # very first batch
    if not recs:
        recs.append({})

    if layer_name in recs[-1]:
        # if the layer is already there, the entry is from the previous training
        # step, thus start a new one
        recs.append({})

    val = to_numpy(output)

    # disambiguate activations when the output is not a simple array
    if isinstance(val, np.ndarray):
        recs[-1][layer_name] = val
    elif isinstance(val, (list, tuple)):
        for i, v in enumerate(val):
            if not isinstance(v, np.ndarray):
                raise TypeError(f"activations of type {type(v)} are not supported")
            recs[-1][layer_name + f'[{i}]'] = v
    elif isinstance(val, Mapping):
        for k, v in val.items():
            if not isinstance(v, np.ndarray):
                raise TypeError(f"Activations of type {type(v)} are not supported")
            recs[-1][layer_name + f'["{k}"]'] = v
    else:
        raise TypeError(f"Activations of type {type(v)} are not supported")


def add_activation_hooks(net, match_fn=None):
    """Add forward hooks to all layers to record activations

    Ignore the top level modules like ``net.module_`` and ``net.criterion_``.

    If an output is not a simple array, it is disambiguated. E.g. if it's a
    list, the name get a suffix of ``[i]`` where ``i`` designates the index in
    the list. Similary, when the output is a dict, a ``[key]`` suffix is added,
    where ``[key]`` is the key of the corresponding value in the dictionary.

    Parameters
    ----------
    net : skorch.NeuralNet
      The neural net instance (has to be initialized).

    match_fn : callable or None (default=None)
      If not ``None``, this should be a callable/function that takes the name of
      a layer as input and returns a bool as output, where ``False`` indicates
      that this layer should be excluded.

    Returns
    -------
    recs : dict of list of dict
      Data structure containing the recorded activations. For each module, for
      each training step, for each layer, there is an entry of the recorded
      activations.

    handles : list of torch.utils.hooks.RemovableHandle
      These handles are returned by torch when adding hooks. They can be used to
      remove the hooks.

    Raises
    ------
    TypeError
      When the output of a layer is not a simple array and if it cannot be
      disambiguated, a ``TypeError`` is raised. Disambiguation works for lists,
      tuples, and dicts (but not nested ones).

    """
    recs = {}
    handles = []

    for module_name, module in named_modules(net):
        recs[module_name] = []
        for layer_name, submodule in module.named_modules():
            if match_fn and not match_fn(layer_name):
                continue

            if submodule is module:
                # is recording activations for whole module useful? skip for now
                continue

            handle = submodule.register_forward_hook(partial(
                _add_activation_hook,
                recs=recs[module_name],
                layer_name=layer_name,
            ))
            handles.append(handle)

    return recs, handles


def _add_grad_hook(grad, *, rec_grad, rec_param_update, param_name, tensor):
    """Helper function for adding gradient hooks"""
    # if the param is already there, the entry is from the previous training
    # step, thus start a new one

    # very first batch
    if not rec_grad:
        rec_grad.append({})
    if not rec_param_update:
        rec_param_update.append({})

    if param_name in rec_grad[-1]:
        rec_grad.append({})
    if param_name in rec_param_update[-1]:
        rec_param_update.append({})

    # We need to clone the grad here because otherwise, the recorded gradients
    # can be overridden by later gradients despite them being pulled to numpy!
    # This only happens on CPU, which suggests it's a strange bug that's perhaps
    # related to caching and/or memory-mapping.
    grad = grad.clone()

    rec_grad[-1][param_name] = to_numpy(grad)

    eps = 1e-9  # prevent divide by 0
    rec_param_update[-1][param_name] = (grad.std() / (eps + tensor.std())).item()


def add_grad_hooks(net, match_fn=None):
    """Add backward hooks to all parameters to record gradients

    Parameters
    ----------
    net : skorch.NeuralNet
      The neural net instance (has to be initialized).

    match_fn : callable or None (default=None)
      If not ``None``, this should be a callable/function that takes the name of
      a parameter as input and returns a bool as output, where ``False``
      indicates that this parameter should be excluded.

    Returns
    -------
    recs_grad : dict of list of dict
      Data structure containing the recorded gradients. For each module, for each
      training step, for each learnable parameter, there is an entry of the
      recorded gradients.

    recs : dict of list of dict
      Data structure containing the recorded parameter updates. For each module,
      for each training step, for each learnable parameter, there is an entry of
      the recorded parameter updates.

    handles : list of torch.utils.hooks.RemovableHandle
      These handles are returned by torch when adding hooks. They can be used to
      remove the hooks.

    """
    recs_grad = {}
    recs_param_update = {}
    handles = []

    for module_name, module in named_modules(net):
        recs_grad[module_name] = []
        recs_param_update[module_name] = []

        for param_name, tensor in module.named_parameters():
            if match_fn and not match_fn(param_name):
                continue

            if not tensor.requires_grad:
                continue

            handle = tensor.register_hook(partial(
                _add_grad_hook,
                rec_grad=recs_grad[module_name],
                rec_param_update=recs_param_update[module_name],
                param_name=param_name,
                tensor=tensor,
            ))
            handles.append(handle)

    return recs_grad, recs_param_update, handles


class SkorchDoctor:
    """SkorchDoctor helps you better understand and debug neural net training.

    By providing records activations, gradienst, and parameter updates, as well
    as useful plotting functions, this class can assist you in understanding
    your training process and how to improve it.

    This is heavily inspred by the tips from Andrej Karpathy on how to better
    understand the neural net and diagnose potential errors.

        - https://karpathy.github.io/2019/04/25/recipe/
        - https://www.youtube.com/watch?v=P6sfmUTpUmc

    To use this class, initialize your skorch net and load your data as you
    would typically do when training a net. Then, initializea ``SkorchDoctor``
    instance by passing said net, and fit the doctor with a small amount of
    data. Finally, use the records or plotting functions to help you better
    understand the training process.

    What exactly you do with this information is up to you. Some examples that
    come to mind:

        - Use the ``plot_loss`` figure to see if your model is powerful enough
          to completely overfit a small sample. If not, consider increasing its
          capacity, e.g. by stacking more layers or using more units per layer.

        - Use the distribution of activations to see if some layers produce
          extreme values and may need a different non-linearity or some form of
          normalization, e.g. batch norm or layer norm. A different weight
          initialization scheme could also help.

        - Check the relative magnitude of the parameter updates to check if your
          learning rate is too low or too high. Maybe some layers should be
          frozen, or you might want to have different learning rates for
          different parameter groups, or an adaptive optimizer like Adam.

        - If gradients are too big, consider using gradient clipping. If the
          mangitude of gradients shifts over time, you might want to use a
          learning rate scheduler.

    At the end of the day, ``SkorchDoctor`` will not tell you what you need to
    do to improve the training process, but it will greatly facilitate to
    diagnose potential problems.

    Examples
    --------
    >>> net = NeuralNet(..., max_epochs=5)  # a couple of epochs are enough
    >>> from skorch.helper import SkorchDoctor
    >>> doctor = SkorchDoctor(net)
    >>> X_sample, y_sample = X[:100], y[:100]  # a few samples are enough
    >>> doctor.fit(X_sample, y_sample)
    >>> # now use the attributes and plotting functions to better
    >>> # understand the training process
    >>> doctor.activation_recs_  # the recorded activations
    >>> doctor.gradient_recs_  # the recorded gradients
    >>> doctor.param_update_recs_  # the recorded parameter updates
    >>> # the next steps require matplotlib to be installed
    >>> doctor.plot_loss()
    >>> doctor.plot_activations()
    >>> doctor.plot_gradients()
    >>> doctor.plot_param_updates()
    >>> doctor.plot_activations_over_time(<layer-name>)
    >>> doctor.plot_gradients_over_time(<param-name>)

    Notes
    -----
    Even if a train/valid split is used for the net, only training data is
    recorded.

    Since ``SkorchDoctor`` will record a lot of values, you should expect an
    increase in memory usage and training time. However, it's sufficient to
    train with a handful of samples and only a few epochs, which helps
    offsetting those disadvantages.

    After you finished the analysis, it is recommended to re-initialize the net
    or even better start a new process. This is because the net you passed is
    modified by adding hooks training it. Although there is a clean up step at
    the end, it's better to start fresh when starting the real model training.

    Parameters
    ----------
    net : skorch.NeuralNet
      The skorch net to be diagnosed.

    match_fn : callable or None (default=None)
      If ``match_fn=None``, all activations, gradients, and parameter updates
      are recorded.

      If not ``None``, this should be a callable/function that takes the name of
      a layer or parameter as input and returns a bool as output, where
      ``False`` indicates that this output should be excluded. As an example, if
      you have a module with a :class:`torch.nn.Linear` layer called ``"fc"``
      and you only want to keep records from that layer, and also not record any
      gradients on biases, the ``match_fn`` could be defined as:

      ``match_fn = lambda name: ("fc" in name) and ("bias" not in name)``

    Attributes
    ----------
    num_steps_ : int
      The total number of training steps. E.g. if trained for 10 epochs with 5
      batches each, that would be 50.

    module_names_ : list of str
      All modules used by the net, typically those are ``"module"`` and
      ``"criterion"``.

    activation_recs_: dict of list of dict of np.ndarray
      The activations of each layer for each module. The outer dict contains one
      entry for each top level module, e.g. ``module`` and ``criterion``. The
      values are lists, one entry for each training step. The entries of those
      lists are again dictionaries, with keys corresponding to the layer name
      and values to the activations of that layer, stored as a numpy array.

      This data structure seems to be a bit complicated at first but its use is
      quite straightforward. E.g. to get the activations of the layer called
      "dense0" of the "module" in epoch 0 and batch 0, use
      ``doctor.activation_recs_['module'][0]['dense0']``.

      If an activation is not a simple array, it is disambiguated. E.g. if it's a
      list, the name get a suffix of ``[i]`` where ``i`` designates the index in
      the list. Similary, when the output is a dict, a ``[key]`` suffix is
      added, where ``[key]`` is the key of the corresponding value in the
      dictionary.

    gradient_recs_: dict of list of dict of np.ndarray
      The gradients of each parameter for each module. The outer dict contains
      one entry for each top level module, e.g. ``module`` and ``criterion``.
      The values are lists, one entry for each training step. The entries of
      those lists are dictionaries, with keys corresponding to the parameter
      name and values to the gradients of that parameter, stored as a numpy
      array. Only learnable parameters are recorded.

      This data structure seems to be a bit complicated at first but its use is
      quite straightforward. E.g. to get the gradient of the parameter called
      "dense0.weight" of the "module" from training step 7, use
      ``doctor.gradient_recs_['module'][7]['dense0.weight]``.

    param_update_recs_: dict of list of dict of float
      The relative parameter update of each parameter for each module. The outer
      dict contains one entry for each top level module, e.g. ``module`` and
      ``criterion``. The values are lists, one entry for each training step. The
      entries of those lists are dictionaries, with keys corresponding to the
      parameter name and values to the standard deviation of the update of that
      parameter, relative to the standard deviation of that parameter itself,
      stored as a float. Only learnable parameters are recorded.

      This data structure seems to be a bit complicated at first but its use is
      quite straightforward. E.g. to get the update of the parameter called
      "dense0.weight" of the "module" in the last training step, use
      ``doctor.paramter_udpate_recs_['module'][-1]['dense0.weight]``.

    fitted_ : bool
      Whether the instance has been fitted.

    """
    def __init__(self, net, match_fn=None):
        self.net = net
        self.match_fn = match_fn

    def initialize(self):
        """Initialize the SkorchDoctor

        This method typically does not need to be invoked explicitly, because it
        is called by ``fit``.

        """
        if not getattr(self.net, 'initialized_', False):
            self.net.initialize()
        self.fitted_ = False

        module_names = [name for name, _ in named_modules(self.net)]
        self.module_names_ = module_names

        activation_recs, activation_handles = add_activation_hooks(
            self.net, self.match_fn
        )
        gradient_recs, param_update_recs, grad_handles = add_grad_hooks(
            self.net, self.match_fn
        )
        self.activation_recs_ = activation_recs
        self.gradient_recs_ = gradient_recs
        self.param_update_recs_ = param_update_recs
        self.handles_ = activation_handles + grad_handles

        if not self.handles_:
            # this means nothing is being recorded
            msg = "No activations, gradients, or updates are being recorded"
            if self.match_fn:
                msg += ", please check the match_fn"
            raise ValueError(msg)

        return self

    def _clean_up(self):
        """Clean up method to be called after training"""
        for handle in self.handles_:
            handle.remove()

    def check_is_fitted(self):
        try:
            sk_check_is_fitted(self, ['fitted_'])
        except NotFittedError as exc:
            msg = (
                f"{self.__class__.__name__} is not initialized yet. Call "
                "'fit(X, y) before using this method."
            )
            raise NotInitializedError(msg) from exc

    def fit(self, X, y=None, **fit_params):
        """Initialize and fit the SkorchDoctor

        It is advised to use a low number of epochs and a small amount of data
        only, since the collection of data results in time and memory overhead.

        The parameters should be exactly the same as those passed when fitting
        the underlying net.

        """
        self.initialize()
        try:
            self.net.partial_fit(X, y, **fit_params)
        finally:
            self._clean_up()

        self.num_steps_ = len(self.activation_recs_[self.module_names_[0]])
        self.fitted_ = True
        return self

    def get_layer_names(self):
        """Return the names of all layers/modules

        Returns
        -------
        names : dict of list of str
          For each top level module, all layer names as a list of strings.

        """
        self.check_is_fitted()
        names = {}
        for module in self.module_names_:
            if self.activation_recs_[module]:
                names[module] = list(self.activation_recs_[module][0].keys())
            else:
                names[module] = []
        return names

    def get_param_names(self):
        """Return all learnable parameters of the net

        Returns
        -------
        names : dict of list of str
          For each top level module, all parameter names as a list of strings.

        """
        self.check_is_fitted()
        names = {}
        for module in self.module_names_:
            if self.gradient_recs_[module]:
                # using the reversed order because gradients are recorded from
                # last to first, but first to last is more intuitive to show
                keys = self.gradient_recs_[module][0].keys()
                names[module] = list(reversed(keys))
            else:
                names[module] = []
        return names

    def predict(self, X, **kwargs):
        """Calls the ``predict`` method of the underlying net"""
        return self.net.predict(X, **kwargs)

    def predict_proba(self, X, **kwargs):
        """Calls the ``predict_proba`` method of the underlying net"""
        return self.net.predict_proba(X, **kwargs)

    def score(self, X, y=None, **kwargs):
        """Calls the ``score`` method of the underlying net"""
        return self.net.score(X, y=y, **kwargs)

    def _get_axes(self, axes=None, figsize=None, nrows=1, squeeze=False):
        """Helper function to get empty axes for plotting

        Unless ``squeeze=True``, a 2-dim array will be returned.

        """
        if axes is not None:
            return axes

        try:
            import matplotlib.pyplot as plt
        except ImportError as exc:
            msg = (
                "This feature requires matplotlib to be installed; "
                "please install it first, e.g. using "
                "'python -m pip install matplotlib'"
            )
            raise ImportError(msg) from exc

        width = 12
        height = 6

        if figsize is None:
            figsize = (width, nrows * height)

        _, axes = plt.subplots(nrows, 1, figsize=figsize, squeeze=squeeze)
        return axes

    def plot_loss(self, ax=None, figsize=None, **kwargs):
        """Plot the loss over each epoch.

        Plots the training loss and, if present, the validation loss over time.

        """
        self.check_is_fitted()

        ax = self._get_axes(ax, figsize=figsize, nrows=1, squeeze=True)
        history = self.net.history
        xvec = np.arange(len(history)) + 1
        ax.plot(xvec, history[:, 'train_loss'], label='train', **kwargs)

        if 'valid_loss' in history[0]:
            ax.plot(xvec, history[:, 'valid_loss'], label='valid', **kwargs)

        ax.set_xlabel('epoch')
        ax.set_ylabel('loss')
        ax.legend()
        ax.set_title('loss over time')
        return ax

    def plot_activations(
            self,
            step=-1,
            match_fn=None,
            axes=None,
            histtype='step',
            lw=2,
            bins=None,
            density=True,
            figsize=None,
            **kwargs
    ):
        """Plot the distribution of activations produced by the layers

        Parameters
        ----------
        step : int (default=-1)
          Which training step to plot. By default, the last step (-1) is chosen.

        match_fn : callable or None (default=None)
          If not ``None``, this should be a callable/function that takes the
          name of a layer as input and returns a bool as output, where ``False``
          indicates that this layer should be excluded. Use this to filter only
          specific layers you want to plot.

        axes : np.ndarray of AxesSubplot or None (default=None)
          By default, a new matplotlib plot is created. If you instead want to
          plot onto an existing plot, pass it here. There should be one subplot
          for each top level module (typically 2).

        bins : np.ndarray or None (default=None)
          Bins to use for the histogram. If left as ``None``, they are inferred
          from the data.

        **kwargs
          You can override remaining plotting arguments like ``lw`` (line width)
          or ``figsize`` (figure size).

        Returns
        -------
        axes : np.andarray of AxesSubplot
          The axes of the plot.

        """
        self.check_is_fitted()

        # only use modules for which the values are not simply empty lists
        module_names = [
            key for key, val in self.activation_recs_.items() if any(l for l in val)
        ]

        axes = self._get_axes(axes, figsize=figsize, nrows=len(module_names))

        for module_name, ax in zip(module_names, axes.flatten()):
            activations = self.activation_recs_[module_name][step]
            if match_fn:
                activations = {k: v for k, v in activations.items() if match_fn(k)}
                if not activations:
                    msg = (
                        "No layer found matching the specification of match_fn. "
                        "Use doctor.get_layer_names() to check all layers."
                    )
                    raise ValueError(msg)

            if bins is None:
                bin_min = min(a.min() for a in activations.values())
                bin_max = max(a.max() for a in activations.values())
                bins = np.linspace(bin_min, bin_max, 50)

            for key, val in activations.items():
                if val.ndim:
                    ax.hist(
                        val.flatten(),
                        label=key,
                        histtype=histtype,
                        lw=lw,
                        bins=bins,
                        density=density,
                        **kwargs
                    )
            ax.legend(loc='best')
            ax.set_title(f"distribution of activations of {module_name}")
        return axes

    def plot_gradients(
            self,
            step=-1,
            match_fn=None,
            axes=None,
            histtype='step',
            lw=2,
            bins=None,
            density=True,
            figsize=None,
            **kwargs
    ):
        """Plot the distribution of gradients of each learnable parameter

        Parameters
        ----------
        step : int (default=-1)
          Which training step to plot. By default, the last step (-1) is chosen.

        match_fn : callable or None (default=None)
          If not ``None``, this should be a callable/function that takes the
          name of a parameter as input and returns a bool as output, where
          ``False`` indicates that this layer should be excluded. Use this to
          filter only specific parameters you want to plot.

        axes : np.ndarray of AxesSubplot or None (default=None)
          By default, a new matplotlib plot is created. If you instead want to
          plot onto an existing plot, pass it here. There should be one subplot
          for each top level module (typically 2).

        bins : np.ndarray or None (default=None)
          Bins to use for the histogram. If left as ``None``, they are inferred
          from the data.

        **kwargs
          You can override remaining plotting arguments like ``lw`` (line width)
          or ``figsize`` (figure size).

        Returns
        -------
        axes : np.andarray of AxesSubplot
          The axes of the plot.

        """
        self.check_is_fitted()

        # only use modules for which the values are not simply empty
        module_names = [
            key for key, val in self.gradient_recs_.items()
            if any(d for l in val for d in l)
        ]

        axes = self._get_axes(axes, figsize=figsize, nrows=len(module_names))

        for module_name, ax in zip(module_names, axes.flatten()):
            gradients = self.gradient_recs_[module_name][step]
            if match_fn:
                gradients = {k: v for k, v in gradients.items() if match_fn(k)}
                if not gradients:
                    msg = (
                        "No parameter found matching the specification of match_fn. "
                        "Use doctor.get_param_names() to check all parameters."
                    )
                    raise ValueError(msg)

            if bins is None:
                bin_min = min(g.min() for g in gradients.values())
                bin_max = max(g.max() for g in gradients.values())
                bins = np.linspace(bin_min, bin_max, 50)

            for key, val in gradients.items():
                if val.ndim:  # don't show scalars
                    ax.hist(
                        val.flatten(),
                        label=key,
                        histtype=histtype,
                        lw=lw,
                        bins=bins,
                        density=density,
                        **kwargs
                    )

            ax.legend(loc='best')
            ax.set_title(f"distribution of gradients for {module_name}")

        return axes

    def plot_param_updates(self, match_fn=None, axes=None, figsize=None, **kwargs):
        """Plot the distribution of relative parameter updates.

        Plots the log10 of the standard deviation of the parameter update
        relative to the parameter itself, over time. Higher values mean that the
        parameter changes quite a lot with each training step, lower values mean
        that the parameter changes little.

        Parameters
        ----------
        match_fn : callable or None (default=None)
          If not ``None``, this should be a callable/function that takes the
          name of a parameter as input and returns a bool as output, where
          ``False`` indicates that this layer should be excluded. Use this to
          filter only specific parameters you want to plot.

        axes : np.ndarray of AxesSubplot or None (default=None)
          By default, a new matplotlib plot is created. If you instead want to
          plot onto an existing plot, pass it here. There should be one subplot
          for each top level module (typically 2).

        bins : np.ndarray or None (default=None)
          Bins to use for the histogram. If left as ``None``, they are inferred
          from the data.

        **kwargs
          You can override remaining plotting arguments like ``figsize`` (figure
          size).

        Returns
        -------
        axes : np.andarray of AxesSubplot
          The axes of the plot.

        """
        self.check_is_fitted()

        # only use modules for which the values are not simply empty
        module_names = [
            key for key, val in self.gradient_recs_.items()
            if any(d for l in val for d in l)
        ]

        axes = self._get_axes(axes, figsize=figsize, nrows=len(module_names))
        eps = 1e-9  # prevent log10 of 0

        for module_name, ax in zip(module_names, axes.flatten()):
            param_updates = self.param_update_recs_[module_name]
            keys = param_updates[0].keys()
            if match_fn:
                keys = [key for key in keys if match_fn(key)]
                if not keys:
                    msg = (
                        "No parameter found matching the specification of match_fn. "
                        "Use doctor.get_param_names() to check all parameters."
                    )
                    raise ValueError(msg)

            for key in keys:
                values = [np.log10(update[key] + eps) for update in param_updates]
                xvec = np.arange(1, len(values) + 1)
                ax.plot(xvec, values, label=key, **kwargs)

            ax.set_xlabel("step")
            ax.set_ylabel(
                f"log10 of stdev of relative parameter updates for {module_name}"
            )
            ax.set_title(module_name)
            ax.legend(loc='best')

        return axes

    def plot_activations_over_time(
            self,
            layer_name,
            module_name='module',
            ax=None,
            lw=2,
            bins=None,
            figsize=None,
            color='k',
            **kwargs
    ):
        """Plot the distribution of the activation of a specific layers over
        time

        The histograms are plotted "on top of each other" with an offset.
        Therefore, the absolute magnitude on the y-axis has no meaning.

        Parameters
        ----------
        layer_name : str
          The name of the specific layer whose activations should be plotted.

        module_name : str (default='module')
          The name of the module that the layer belongs to. By default, it is
          called "module" in skorch, but it's possible to define custom module
          names, in which case the corresponding name should be chosen.

        ax : AxesSubplot or None (default=None)
          By default, a new matplotlib plot is created. If you instead want to
          plot onto an existing plot, pass it here. Only a single plot is
          created.

        bins : np.ndarray or None (default=None)
          Bins to use for the histogram. If left as ``None``, they are inferred
          from the data.

        **kwargs
          You can override remaining plotting arguments like ``figsize`` (figure
          size).

        Returns
        -------
        ax : AxesSubplot
          The ax of the plot.

        """
        self.check_is_fitted()

        ax = self._get_axes(ax, figsize=figsize, squeeze=True)

        try:
            activations = [
                act[layer_name] for act in self.activation_recs_[module_name]
            ]
        except KeyError as exc:
            msg = (
                f"No layer named '{layer_name}' could be found. "
                "Use doctor.get_layer_names() to check all layers."
            )
            raise ValueError(msg) from exc

        n = len(activations)

        if bins is None:
            bin_min = min(a.min() for a in activations)
            bin_max = max(a.max() for a in activations)
            bins = np.linspace(bin_min, bin_max, 100)

        y_max = -1
        yvals = []
        for activation in activations:
            yval, _ = np.histogram(activation.flatten(), bins=bins)
            y_max = max(y_max, yval.max())
            yvals.append(yval)

        y_scale = y_max / n / 2
        alpha = max(0.1, 1 / n)

        for step, yval in enumerate(reversed(yvals)):
            bottom = y_scale * (n - step - 1)
            ax.fill_between(
                bins[1:],
                yval + bottom,
                y2=bottom,
                alpha=alpha,
                color=color,
                lw=lw,
                **kwargs
            )

        ax.set_yticklabels([])
        ax.set_ylabel('step / activation')
        ax.set_title(f"distribution of activations for {layer_name} of {module_name}")
        return ax

    def plot_gradient_over_time(
            self,
            param_name,
            module_name='module',
            ax=None,
            lw=2,
            bins=None,
            figsize=None,
            color='k',
            **kwargs
    ):
        """Plot the distribution of the gradients of a specific parameter over
        time

        The histograms are plotted "on top of each other" with an offset.
        Therefore, the absolute magnitude on the y-axis has no meaning.

        Parameters
        ----------
        param_name : str
          The name of the specific parameter that should be plotted.

        module_name : str (default='module')
          The name of the module that the paramter belongs to. By default, it is
          called "module" in skorch, but it's possible to define custom module
          names, in which case the corresponding name should be chosen.

        ax : AxesSubplot or None (default=None)
          By default, a new matplotlib plot is created. If you instead want to
          plot onto an existing plot, pass it here. Only a single plot is
          created.

        bins : np.ndarray or None (default=None)
          Bins to use for the histogram. If left as ``None``, they are inferred
          from the data.

        **kwargs
          You can override remaining plotting arguments like ``figsize`` (figure
          size).

        Returns
        -------
        ax : AxesSubplot
          The ax of the plot.

        """
        self.check_is_fitted()

        ax = self._get_axes(ax, figsize=figsize, squeeze=True)

        try:
            gradients = [
                grad[param_name] for grad in self.gradient_recs_[module_name]
            ]
        except KeyError as exc:
            msg = (
                f"No parameter named '{param_name}' could be found. "
                "Use doctor.get_param_names() to check all parameters."
            )
            raise ValueError(msg) from exc


        n = len(gradients)

        if bins is None:
            bin_min = min(g.min() for g in gradients)
            bin_max = max(g.max() for g in gradients)
            bins = np.linspace(bin_min, bin_max, 100)

        y_max = -1
        yvals = []
        for gradient in gradients:
            yval, _ = np.histogram(gradient.flatten(), bins=bins)
            y_max = max(y_max, yval.max())
            yvals.append(yval)

        y_scale = y_max / n / 2
        alpha = max(0.1, 1 / n)

        for step, yval in enumerate(reversed(yvals)):
            bottom = y_scale * (n - step - 1)
            ax.fill_between(
                bins[1:],
                yval + bottom,
                y2=bottom,
                alpha=alpha,
                color=color,
                lw=lw,
                **kwargs)

        ax.set_yticklabels([])
        ax.set_ylabel('step / gradient')
        ax.set_title(f"distribution of gradients for {param_name} of {module_name}")
        return ax



"""Tests for skorch._doctor.py"""

import itertools
from unittest import mock

import numpy as np
import pytest
import torch
from torch import nn


class TestSkorchDoctorSimple:  # pylint: disable=too-many-public-methods
    """Test functionality of SkorchDoctor using a simple model"""
    @pytest.fixture(scope='module')
    def module_cls(self):
        """Return a simple module class with predictable parameters"""
        class MyModule(nn.Module):
            """Module with predictable parameters"""
            def __init__(self):
                super().__init__()

                self.lin0 = nn.Linear(20, 20)
                nn.init.eye_(self.lin0.weight)
                nn.init.zeros_(self.lin0.bias)

                self.lin1 = nn.Linear(20, 2)
                nn.init.zeros_(self.lin1.weight)
                nn.init.ones_(self.lin1.bias)

                self.softmax = nn.Softmax(dim=-1)

            def forward(self, X):
                X = self.lin0(X)
                X = self.lin1(X)
                return self.softmax(X)

        return MyModule

    @pytest.fixture(scope='module')
    def custom_split(self):
        """Split train/valid deterministically so that we know all training
        samples"""
        class Split():
            """Deterministically split train/valid into 80%/20%"""
            def __call__(self, dataset, y=None, groups=None):
                n = int(len(dataset) * 0.8)
                dataset_train = torch.utils.data.Subset(dataset, np.arange(n))
                dataset_valid = torch.utils.data.Subset(
                    dataset, np.arange(n, len(dataset)))
                return dataset_train, dataset_valid

        return Split()

    @pytest.fixture(scope='module')
    def net_cls(self):
        from skorch import NeuralNetClassifier
        return NeuralNetClassifier

    @pytest.fixture(scope='module')
    def doctor_cls(self):
        from skorch.helper import SkorchDoctor
        return SkorchDoctor

    @pytest.fixture(scope='module')
    def data(self, classifier_data):
        X, y = classifier_data
        # a small amount of data is enough
        return X[:50], y[:50]

    @pytest.fixture(scope='module')
    def doctor(self, module_cls, net_cls, doctor_cls, data, custom_split):
        net = net_cls(module_cls, max_epochs=3, batch_size=32, train_split=custom_split)
        doctor = doctor_cls(net)
        doctor.fit(*data)
        return doctor

    def test_activation_recs_general_content(self, doctor):
        recs = doctor.activation_recs_
        assert set(recs.keys()) == {'module', 'criterion'}

        # nothing recorded for criterion
        assert recs['criterion'] == []

        recs_module = recs['module']
        # 3 epochs, 2 batches per epoch
        assert len(recs_module) == 6

        # each batch has layers lin0, lin1, softmax
        for batch in recs_module:
            assert set(batch.keys()) == {'lin0', 'lin1', 'softmax'}

    def test_activation_recs_values(self, doctor, data):
        recs_module = doctor.activation_recs_['module']

        for key in ('lin0', 'lin1', 'softmax'):
            # 80% of 50 samples is 40, batch size 32 => 32 + 8 samples per batch
            batch_sizes = [len(batch[key]) for batch in recs_module]
            assert batch_sizes == [32, 8, 32, 8, 32, 8]

        X, _ = data
        # for the very first batch, before any update, we actually know the values
        batch0 = recs_module[0]
        lin0_0 = batch0['lin0']
        # since it is the identity function, batches should equal the data
        np.testing.assert_array_almost_equal(lin0_0, X[:32])

        lin1_0 = batch0['lin1']
        # since weights are 0 and bias is 1, all values should be 1
        np.testing.assert_array_almost_equal(lin1_0, 1.0)

        softmax_0 = batch0['softmax']
        # since all inputs are equal, probabilities should be uniform
        np.testing.assert_array_almost_equal(softmax_0, 0.5)

    def test_activation_recs_not_all_identical(self, doctor):
        # make sure that values are not just all identical, using a large
        # tolerance to exclude small deviations
        recs = doctor.activation_recs_['module']

        recs_lin0 = [rec['lin0'] for rec in recs]
        for act0, act1 in itertools.combinations(recs_lin0, r=2):
            if act0.shape == act1.shape:
                assert not np.allclose(act0, act1, rtol=1e-3)

        recs_lin1 = [rec['lin1'] for rec in recs]
        for act0, act1 in itertools.combinations(recs_lin1, r=2):
            if act0.shape == act1.shape:
                assert not np.allclose(act0, act1, rtol=1e-3)

        softmax = [rec['softmax'] for rec in recs]
        for act0, act1 in itertools.combinations(softmax, r=2):
            if act0.shape == act1.shape:
                assert not np.allclose(act0, act1, rtol=1e-3)

    def test_gradient_recs_general_content(self, doctor):
        recs = doctor.gradient_recs_
        assert set(recs.keys()) == {'module', 'criterion'}

        # nothing recorded for criterion, 3 epochs
        assert recs['criterion'] == []

        recs_module = recs['module']
        # 3 epochs, 2 batches per epoch
        assert len(recs_module) == 6

        # each batch has weights and biases for lin0 & lin1
        expected = {'lin0.weight', 'lin0.bias', 'lin1.weight', 'lin1.bias'}
        for batch in recs_module:
            assert set(batch.keys()) == expected

    def test_gradient_recs_values(self, doctor):
        recs_module = doctor.gradient_recs_['module']

        expected_shapes = {
            'lin0.weight': (20, 20),
            'lin0.bias': (20,),
            'lin1.weight': (2, 20),
            'lin1.bias': (2,),
        }
        for key in ('lin0.weight', 'lin0.bias', 'lin1.weight', 'lin1.bias'):
            grad_shapes = [batch[key].shape for batch in recs_module]
            expected_shape = expected_shapes[key]
            # 2 batches, 3 epochs
            assert grad_shapes == [expected_shape] * 6

        # There is not really much we know about the gradient values, we just
        # rely on the gradient hooks doing what they're supposed to do. The only
        # gradients we actually know are for the first layer in the first batch:
        # They have to be zero because in the second layer, we have weights of 0.

        batch0 = recs_module[0]
        grad_weight = batch0['lin0.weight']
        grad_bias = batch0['lin0.bias']
        assert np.allclose(grad_weight, 0.0)
        assert np.allclose(grad_bias, 0.0)

    def test_gradient_recs_not_all_identical(self, doctor):
        # make sure that values are not just all identical, using a large
        # tolerance to exclude small deviations
        recs = doctor.gradient_recs_['module']

        recs_lin0_weight = [rec['lin0.weight'] for rec in recs]
        for grad0, grad1 in itertools.combinations(recs_lin0_weight, r=2):
            assert not np.allclose(grad0, grad1, rtol=1e-3)

        recs_lin0_bias = [rec['lin0.bias'] for rec in recs]
        for grad0, grad1 in itertools.combinations(recs_lin0_bias, r=2):
            assert not np.allclose(grad0, grad1, rtol=1e-3)

        recs_lin1_weight = [rec['lin1.weight'] for rec in recs]
        for grad0, grad1 in itertools.combinations(recs_lin1_weight, r=2):
            assert not np.allclose(grad0, grad1, rtol=1e-3)

        recs_lin1_bias = [rec['lin1.bias'] for rec in recs]
        for grad0, grad1 in itertools.combinations(recs_lin1_bias, r=2):
            assert not np.allclose(grad0, grad1, rtol=1e-3)

    def test_param_update_recs_general_content(self, doctor):
        recs = doctor.param_update_recs_
        assert set(recs.keys()) == {'module', 'criterion'}

        # nothing recorded for criterion
        assert recs['criterion'] == []

        recs_module = recs['module']
        # 3 epochs, 2 batches per epoch
        assert len(recs_module) == 6

        # each batch has weights and biases for lin0 & lin1
        expected = {'lin0.weight', 'lin0.bias', 'lin1.weight', 'lin1.bias'}
        for batch in recs_module:
            assert set(batch.keys()) == expected

    def test_param_update_recs_values(self, doctor):
        recs= doctor.param_update_recs_['module']
        assert all(np.isscalar(val) for d in recs for val in d.values())

        # for the very first batch, before any update, we actually know that the
        # updates must be 0 because the gradients are 0.
        batch0 = recs[0]
        assert np.isclose(batch0['lin0.weight'], 0)
        assert np.isclose(batch0['lin0.bias'], 0)

    def test_param_update_recs_not_all_identical(self, doctor):
        # make sure that values are not just all identical, using a large
        # tolerance to exclude small deviations
        recs = doctor.param_update_recs_['module']

        recs_lin0_weight = [rec['lin0.weight'] for rec in recs]
        for upd0, upd1 in itertools.combinations(recs_lin0_weight, r=2):
            assert not np.isclose(upd0, upd1, rtol=1e-3)

        recs_lin0_bias = [rec['lin0.bias'] for rec in recs]
        for upd0, upd1 in itertools.combinations(recs_lin0_bias, r=2):
            assert not np.isclose(upd0, upd1, rtol=1e-3)

        recs_lin1_weight = [rec['lin1.weight'] for rec in recs]
        for upd0, upd1 in itertools.combinations(recs_lin1_weight, r=2):
            assert not np.isclose(upd0, upd1, rtol=1e-3)

        recs_lin1_bias = [rec['lin1.bias'] for rec in recs]
        for upd0, upd1 in itertools.combinations(recs_lin1_bias, r=2):
            assert not np.isclose(upd0, upd1, rtol=1e-3)

    def test_hooks_cleaned_up_after_fit(self, doctor, data):
        # make sure that the hooks are cleaned up by checking that no more recs
        # are written when continuing to fit the net
        num_activation_recs_before = len(doctor.activation_recs_['module'])
        num_gradient_recs_before = len(doctor.gradient_recs_['module'])

        net = doctor.net
        net.partial_fit(*data)

        num_activation_recs_after = len(doctor.activation_recs_['module'])
        num_gradient_recs_after = len(doctor.gradient_recs_['module'])

        assert num_activation_recs_before == num_activation_recs_after
        assert num_gradient_recs_before == num_gradient_recs_after

    def test_callbacks_cleaned_up_after_fit_with_initial_callbacks(
            self, doctor_cls, net_cls, module_cls, data
    ):
        # make sure that the callbacks are the same before and after, this is
        # important because SkorchDoctor will temporarily add a callback
        from skorch.callbacks import EpochScoring, GradientNormClipping

        net = net_cls(
            module_cls,
            callbacks=[EpochScoring('f1'), GradientNormClipping(1.0)],
        ).initialize()
        callbacks_without_doctor = net.callbacks_[:]

        doctor = doctor_cls(net).fit(*data)
        callbacks_with_doctor = doctor.net.callbacks_

        assert len(callbacks_without_doctor) == len(callbacks_with_doctor)

        for (name0, cb0), (name1, cb1) in zip(
                callbacks_without_doctor, callbacks_with_doctor
        ):
            assert name0 == name1
            # pylint: disable=unidiomatic-typecheck
            assert type(cb0) == type(cb1)

    def test_get_layer_names(self, doctor):
        layer_names = doctor.get_layer_names()
        expected = {
            'criterion': [],
            'module': ['lin0', 'lin1', 'softmax']
        }
        assert layer_names == expected

    def test_get_parameter_names(self, doctor):
        param_names = doctor.get_param_names()
        expected = {
            'criterion': [],
            'module': ['lin0.weight', 'lin0.bias', 'lin1.weight', 'lin1.bias'],
        }
        assert param_names == expected

    def test_predict(self, doctor, data):
        X, _ = data
        y_pred_doctor = doctor.predict(X)
        y_pred_net = doctor.net.predict(X)
        np.testing.assert_allclose(y_pred_doctor, y_pred_net)

    def test_predict_proba(self, doctor, data):
        X, _ = data
        y_proba_doctor = doctor.predict_proba(X)
        y_proba_net = doctor.net.predict_proba(X)
        np.testing.assert_allclose(y_proba_doctor, y_proba_net)

    def test_score(self, doctor, data):
        X, y = data
        score_doctor = doctor.score(X, y)
        score_net = doctor.net.score(X, y)
        np.testing.assert_allclose(score_doctor, score_net)

    def test_recs_with_filter(self, module_cls, net_cls, doctor_cls, data, custom_split):
        # when initializing SkorchDoctor with a match_fn, only records whose
        # keys match should be kept
        def match_fn(name):
            return "lin0" in name

        net = net_cls(module_cls, max_epochs=3, batch_size=32, train_split=custom_split)
        doctor = doctor_cls(net, match_fn=match_fn)
        doctor.fit(*data)

        # check trivial case of empty lists
        assert doctor.activation_recs_['module']
        assert doctor.gradient_recs_['module']
        assert doctor.param_update_recs_['module']

        for rec in doctor.activation_recs_['module']:
            for key in rec.keys():
                assert match_fn(key)

        for rec in doctor.gradient_recs_['module']:
            for key in rec.keys():
                assert match_fn(key)

        for rec in doctor.param_update_recs_['module']:
            for key in rec.keys():
                assert match_fn(key)

    def test_recs_with_filter_no_match(
            self, module_cls, net_cls, doctor_cls, data, custom_split
    ):
        # raise a helpful error if the match function filters away everything
        def match_fn(name):
            return "this-substring-does-not-exist" in name

        net = net_cls(module_cls, max_epochs=3, batch_size=32, train_split=custom_split)
        doctor = doctor_cls(net, match_fn=match_fn)

        msg = (
            "No activations, gradients, or updates are being recorded, "
            "please check the match_fn"
        )
        with pytest.raises(ValueError, match=msg):
            doctor.fit(*data)

    ############
    # PLOTTING #
    ############

    # Just do very basic plotting tests, not exact content, just that it works

    @pytest.fixture
    def mock_matplotlib_not_installed(self):
        # fixture to make it seem like matplotlib was not installed
        orig_import = __import__

        def import_mock(name, *args):
            if name == 'matplotlib':
                # pretend that matplotlib is not installed
                raise ModuleNotFoundError("no module named 'matplotlib'")
            return orig_import(name, *args)

        with mock.patch('builtins.__import__', side_effect=import_mock):
            yield import_mock

    # pylint: disable=unused-argument
    def test_matplotlib_not_installed(self, mock_matplotlib_not_installed, doctor):
        # Note: Unfortunately, the order of tests matters here: This test should
        # run before the ones below that use matplotlib, otherwise the import
        # mock doesn't work correctly.
        msg = (
            r"This feature requires matplotlib to be installed; "
            r"please install it first, e.g. using "
            r"\'python -m pip install matplotlib\'"
        )
        with pytest.raises(ImportError, match=msg):
            doctor.plot_loss()

    @pytest.fixture(scope='module')
    def plt(self):
        matplotlib = pytest.importorskip('matplotlib')
        matplotlib.use("agg")

        import matplotlib.pyplot as plt
        # not sure why closing is important but sklearn does it:
        # https://github.com/scikit-learn/scikit-learn/blob/964189df31dd2aa037c5bc58c96f88193f61253b/sklearn/conftest.py#L193
        plt.close("all")
        yield plt
        plt.close("all")

    # pylint: disable=unused-argument
    def test_plot_not_fitted_raises(self, plt, doctor_cls, net_cls, module_cls):
        # testing only one of the plotting functions, but all support it
        from skorch.exceptions import NotInitializedError

        doctor = doctor_cls(net_cls(module_cls))
        msg = (
            r"SkorchDoctor is not initialized yet. Call 'fit\(X, y\) before using this "
            "method."
        )
        with pytest.raises(NotInitializedError, match=msg):
            doctor.plot_loss()

    def test_plot_loss_default(self, plt, doctor):
        ax = doctor.plot_loss()
        assert isinstance(ax, plt.Subplot)

    def test_plot_loss_non_default(self, plt, doctor):
        _, ax = plt.subplots()
        ax_after = doctor.plot_loss(ax=ax, figsize=(1, 2), lw=5)
        assert isinstance(ax, plt.Subplot)
        assert ax_after is ax

    def test_plot_activations_default(self, plt, doctor):
        axes = doctor.plot_activations()
        assert isinstance(axes, np.ndarray)
        assert axes.shape == (1, 1)
        assert isinstance(axes[0, 0], plt.Subplot)

    def test_plot_activations_non_default(self, plt, doctor):
        _, axes = plt.subplots(1, 1, squeeze=False)
        axes_after = doctor.plot_activations(
            axes=axes,
            step=0,
            match_fn=lambda name: 'lin' in name,
            histtype='bar',
            lw=3,
            bins=np.arange(10),
            density=False,
            figsize=(1, 2),
            align='left',
        )
        assert axes_after is axes

    def test_plot_activations_passed_axes_2d(self, plt, doctor):
        _, axes = plt.subplots(2, 2)
        axes_after = doctor.plot_activations(
            axes=axes,
            step=0,
            match_fn=lambda name: 'lin' in name,
            histtype='bar',
            lw=3,
            bins=np.arange(10),
            density=False,
            figsize=(1, 2),
            align='left',
        )
        assert axes_after is axes

    # pylint: disable=unused-argument
    def test_plot_activations_no_match(self, plt, doctor):
        msg = (
            r"No layer found matching the specification of match_fn. "
            r"Use doctor.get_layer_names\(\) to check all layers."
        )

        with pytest.raises(ValueError, match=msg):
            doctor.plot_activations(match_fn=lambda name: 'foo' in name)

    def test_plot_gradients_default(self, plt, doctor):
        axes = doctor.plot_gradients()
        assert isinstance(axes, np.ndarray)
        assert axes.shape == (1, 1)
        assert isinstance(axes[0, 0], plt.Subplot)

    def test_plot_gradients_non_default(self, plt, doctor):
        _, axes = plt.subplots(1, 1, squeeze=False)
        axes_after = doctor.plot_gradients(
            axes=axes,
            step=0,
            match_fn=lambda name: 'lin' in name,
            histtype='bar',
            lw=3,
            bins=np.arange(10),
            density=False,
            figsize=(1, 2),
            align='left',
        )
        assert axes_after is axes

    # pylint: disable=unused-argument
    def test_plot_gradients_no_match(self, plt, doctor):
        msg = (
            r"No parameter found matching the specification of match_fn. "
            r"Use doctor.get_param_names\(\) to check all parameters."
        )

        with pytest.raises(ValueError, match=msg):
            doctor.plot_gradients(match_fn=lambda name: 'foo' in name)

    def test_plot_param_updates_default(self, plt, doctor):
        axes = doctor.plot_param_updates()
        assert isinstance(axes, np.ndarray)
        assert axes.shape == (1, 1)
        assert isinstance(axes[0, 0], plt.Subplot)

    def test_plot_param_updates_non_default(self, plt, doctor):
        _, axes = plt.subplots(1, 1, squeeze=False)
        axes_after = doctor.plot_param_updates(
            axes=axes,
            match_fn=lambda name: 'lin' in name,
            lw=3,
            figsize=(1, 2),
        )
        assert isinstance(axes, np.ndarray)
        assert axes_after is axes

    # pylint: disable=unused-argument
    def test_plot_param_updates_no_match(self, plt, doctor):
        msg = (
            r"No parameter found matching the specification of match_fn. "
            r"Use doctor.get_param_names\(\) to check all parameters."
        )

        with pytest.raises(ValueError, match=msg):
            doctor.plot_param_updates(match_fn=lambda name: 'foo' in name)

    def test_plot_activations_over_time_default(self, plt, doctor):
        ax = doctor.plot_activations_over_time(layer_name='lin0')
        assert isinstance(ax, plt.Subplot)

    def test_plot_activations_over_time_non_default(self, plt, doctor):
        _, ax = plt.subplots()
        ax_after = doctor.plot_activations_over_time(
            layer_name='softmax',
            ax=ax,
            lw=3,
            bins=np.arange(10),
            color='r',
            figsize=(1, 2),
            interpolate=True,
        )
        assert ax_after is ax

    # pylint: disable=unused-argument
    def test_plot_activations_over_time_no_match(self, plt, doctor):
        msg = (
            r"No layer named 'foo' could be found. "
            r"Use doctor.get_layer_names\(\) to check all layers."
        )

        with pytest.raises(ValueError, match=msg):
            doctor.plot_activations_over_time(layer_name='foo')

    def test_plot_gradient_over_time_default(self, plt, doctor):
        ax = doctor.plot_gradient_over_time(param_name='lin1.weight')
        assert isinstance(ax, plt.Subplot)

    def test_plot_gradient_over_time_non_default(self, plt, doctor):
        _, ax = plt.subplots()
        ax_after = doctor.plot_gradient_over_time(
            param_name='lin0.bias',
            ax=ax,
            lw=3,
            bins=np.arange(10),
            color='r',
            figsize=(1, 2),
            interpolate=True,
        )
        assert ax_after is ax

    # pylint: disable=unused-argument
    def test_plot_gradient_over_time_no_match(self, plt, doctor):
        msg = (
            r"No parameter named 'foo' could be found. "
            r"Use doctor.get_param_names\(\) to check all parameters."
        )

        with pytest.raises(ValueError, match=msg):
            doctor.plot_gradient_over_time(param_name='foo')


class TestSkorchDoctorComplexArchitecture:
    """Tests based on a more non-standard model

    Specifically, add modules with non-standard names and non-standard outputs
    like tuples or dicts.

    This test class does not re-iterate all the tests performed on the standard
    model but focuses on the parts that change, e.g. how the outputs are
    recorded.

    """
    @pytest.fixture(scope='module')
    def module0_cls(self):
        """Module that returns a tuple"""
        class MyModule(nn.Module):
            """Module that returns a tuple, lin0 no grad"""
            def __init__(self):
                super().__init__()

                self.lin0 = nn.Linear(20, 20)
                self.lin0.requires_grad_(False)
                self.lin1 = nn.Linear(20, 2)

            def forward(self, X):
                X0 = self.lin0(X)
                X1 = self.lin1(X)
                return X0, X1

        return MyModule

    @pytest.fixture(scope='module')
    def module1_cls(self):
        """Module without learnable params that returns a dict"""
        class MyModule(nn.Module):
            """Module that returns a dict"""
            def __init__(self):
                super().__init__()
                self.softmax = nn.Softmax(dim=-1)

            def forward(self, X):
                softmax = self.softmax(X)
                return {'logits': X, 'softmax': softmax}

        return MyModule

    @pytest.fixture(scope='module')
    def module2_cls(self, module0_cls, module1_cls):
        """Module that combines module0 and module1"""
        class MyModule(nn.Module):
            """Module that returns a dict"""
            def __init__(self):
                super().__init__()
                self.module0 = module0_cls()
                self.module1 = module1_cls()

            def forward(self, X):
                _, X1 = self.module0(X)
                output = self.module1(X1)
                return output['softmax']

        return MyModule

    @pytest.fixture(scope='module')
    def criterion_cls(self):
        class MyCriterion(nn.Module):
            """Criterion that has learnable parameters"""
            def __init__(self):
                super().__init__()
                self.lin0 = nn.Linear(2, 2)

            # pylint: disable=arguments-differ
            def forward(self, y_proba, y):
                y_proba = self.lin0(y_proba)
                return nn.functional.nll_loss(y_proba, y)

        return MyCriterion

    @pytest.fixture(scope='module')
    def net_cls(self, module2_cls):
        """Customize net to work with complex modules"""
        from skorch import NeuralNetClassifier
        from skorch.utils import to_tensor

        class MyNet(NeuralNetClassifier):
            """Customize net that works with non-standard modules"""
            def initialize_module(self):
                kwargs = self.get_params_for('mymodule')
                module = self.initialized_instance(module2_cls, kwargs)
                # pylint: disable=attribute-defined-outside-init
                self.mymodule_ = module
                self.seq_ = nn.Sequential(nn.Linear(2, 2))

                return self

            def initialize_criterion(self):
                # use non-standard name 'mycriterion'
                kwargs = self.get_params_for('mycriterion')
                criterion = self.initialized_instance(self.criterion, kwargs)
                # pylint: disable=attribute-defined-outside-init
                self.mycriterion_ = criterion
                return self

            def infer(self, x, **fit_params):
                x = to_tensor(x, device=self.device)
                x = self.mymodule_(x, **fit_params)
                x = x + self.seq_(x)
                return x

            def get_loss(self, y_pred, y_true, *args, **kwargs):
                y_true = to_tensor(y_true, device=self.device)
                return self.mycriterion_(y_pred, y_true)

        return MyNet

    @pytest.fixture(scope='module')
    def doctor_cls(self):
        from skorch.helper import SkorchDoctor
        return SkorchDoctor

    @pytest.fixture(scope='module')
    def data(self, classifier_data):
        X, y = classifier_data
        # a small amount of data is enough
        return X[:50], y[:50]

    @pytest.fixture(scope='module')
    def doctor(self, module0_cls, criterion_cls, net_cls, doctor_cls, data):
        # the passed module doesn't matter as they are hard-coded
        torch.manual_seed(0)
        net = net_cls(
            module0_cls, criterion=criterion_cls, max_epochs=3, batch_size=32
        )
        doctor = doctor_cls(net)
        doctor.fit(*data)
        return doctor

    def test_activation_recs_general_content(self, doctor):
        recs = doctor.activation_recs_
        assert set(recs.keys()) == {'mymodule', 'seq', 'mycriterion'}

        for rec in recs.values():
            # 3 epochs, 2 batches per epoch
            assert len(rec) == 6

        expected_mymodule = {
            'module0.lin0', 'module0.lin1', 'module0[0]', 'module0[1]',
            'module1.softmax', 'module1["logits"]', 'module1["softmax"]',
        }
        assert set(recs['mymodule'][0].keys()) == expected_mymodule
        # nn.Sequential just enumerates the layers
        assert set(recs['seq'][0].keys()) == {'0'}
        assert set(recs['mycriterion'][0].keys()) == {'lin0'}

    def test_gradient_recs_general_content(self, doctor):
        recs = doctor.gradient_recs_
        assert len(recs) == 3
        assert set(recs.keys()) == {'mymodule', 'seq', 'mycriterion'}

        for rec in recs.values():
            # 3 epochs, 2 batches per epoch
            assert len(rec) == 6

        # each batch has weights and biases for lin1, lin0 has no gradient
        expected = {'module0.lin1.weight', 'module0.lin1.bias'}
        for batch in recs['mymodule']:
            assert set(batch.keys()) == expected

        # each batch has weights and biases for lin1, lin0 has no gradient
        expected = {'0.weight', '0.bias'}
        for batch in recs['seq']:
            assert set(batch.keys()) == expected

        # each batch has weights and biases for lin1, lin0 has no gradient
        expected = {'lin0.weight', 'lin0.bias'}
        for batch in recs['mycriterion']:
            assert set(batch.keys()) == expected

    def test_get_layer_names(self, doctor):
        layer_names = doctor.get_layer_names()
        expected = {
            'mymodule': [
                'module0.lin0', 'module0.lin1', 'module0[0]', 'module0[1]',
                'module1.softmax', 'module1["logits"]', 'module1["softmax"]',
            ],
            'seq': ['0'],
            'mycriterion': ['lin0'],
        }
        assert layer_names == expected

    def test_get_parameter_names(self, doctor):
        param_names = doctor.get_param_names()
        expected = {
            'mymodule': ['module0.lin1.weight', 'module0.lin1.bias'],
            'seq': ['0.weight', '0.bias'],
            'mycriterion': ['lin0.weight', 'lin0.bias'],
        }
        assert param_names == expected



from skorch.llm.classifier import FewShotClassifier, ZeroShotClassifier
from skorch.llm.prompts import DEFAULT_PROMPT_FEW_SHOT, DEFAULT_PROMPT_ZERO_SHOT


__all__ = [
    'DEFAULT_PROMPT_FEW_SHOT',
    'DEFAULT_PROMPT_ZERO_SHOT',
    'FewShotClassifier',
    'ZeroShotClassifier',
]
